import asyncio
import aiofiles
import struct
from .enums import TaskScheduleStatus, TaskExecutionStatus, TaskExecutionStatus, WorkerPingReply
from .exceptions import NotEnoughResources, ProcessInitializationError, WorkerNotAvailable, AlreadyRunning
from .environment_resolver import ResolutionImpossibleError
from . import logging
from . import invocationjob
from . import nethelpers

import os

from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from .worker import Worker


class WorkerTaskServerProtocol(asyncio.StreamReaderProtocol):

    def __init__(self, worker: "Worker", limit: int = 2 ** 16):
        self.__logger = logging.get_logger('worker.protocol')
        self.__timeout = 300.0
        self.__reader = asyncio.StreamReader(limit=limit)
        self.__worker = worker
        self.__saved_references = []
        super(WorkerTaskServerProtocol, self).__init__(self.__reader, self.client_connected_cb)

    async def client_connected_cb(self, reader, writer):
        # there is a bug in py <=3.8, callback task can be GCd
        # see https://bugs.python.org/issue46309
        # so we HAVE to save a reference to self somewhere
        self.__saved_references.append(asyncio.current_task())

        async def read_string() -> str:
            strlen = struct.unpack('>Q', await reader.readexactly(8))[0]
            if strlen == 0:
                return ''
            return (await reader.readexactly(strlen)).decode('UTF-8')

        async def write_string(s: str):
            b = s.encode('UTF-8')
            writer.write(struct.pack('>Q', len(b)))
            writer.write(b)

        try:
            await self.__worker.wait_till_starts()  # important that worker is fully up before we actually start listening to shit

            prot = await asyncio.wait_for(reader.readexactly(4), timeout=self.__timeout)
            if prot == b'\0\0\0\0':
                # TODO: either we need timeout here - then add it to all reading operations below
                # TODO: or we don't need it here - then remove wait.
                command = await asyncio.wait_for(reader.readline(), timeout=self.__timeout)  # type: bytes
                if command.endswith(b'\n'):
                    command = command[:-1]
                self.__logger.debug(f'got command: {command.decode("UTF-8")}')

                #
                # command ping
                if command == b'ping':  # routine ping, report worker status
                    if self.__worker.is_task_running():
                        pstatus = WorkerPingReply.BUSY.value
                        pvalue = int(self.__worker.task_status() or 0)
                    else:
                        pstatus = WorkerPingReply.IDLE.value
                        pvalue = 0
                    writer.write(struct.pack('>Bl', pstatus, pvalue))
                #
                # command enqueue task
                elif command == b'task':  # scheduler wants us to pick up task
                    raise DeprecationWarning('this does not work any more')  # TODO: slavage if any and clean up
                    tasksize = await reader.readexactly(4)
                    tasksize = struct.unpack('>I', tasksize)[0]
                    task = await reader.readexactly(tasksize)
                    addrsize = await reader.readexactly(4)
                    addrsize = struct.unpack('>I', addrsize)[0]
                    addr = await reader.readexactly(addrsize)
                    task = invocationjob.InvocationJob.deserialize(task)
                    addr = addr.decode('UTF-8')
                    self.__logger.debug(f'got task: {task}, reply result to {addr}')
                    try:
                        self.__logger.debug('taking the task')
                        await self.__worker.run_task(task, addr)
                        writer.write(bytes([TaskScheduleStatus.SUCCESS.value]))
                    except AlreadyRunning:
                        self.__logger.debug('BUSY. rejecting task')
                        writer.write(bytes([TaskScheduleStatus.BUSY.value]))
                    except ResolutionImpossibleError:
                        self.__logger.info('Worker failed to resolve required environment. rejecting task')
                        writer.write(bytes([TaskScheduleStatus.FAILED.value]))
                    except ProcessInitializationError:
                        self.__logger.info('Failed to initialize payload process. rejecting task')
                        writer.write(bytes([TaskScheduleStatus.FAILED.value]))
                    except NotEnoughResources:
                        self.__logger.warning('Not enough resources (this is unusual error - scheduler should know our resources). rejecting task')
                        writer.write(bytes([TaskScheduleStatus.FAILED.value]))
                    except WorkerNotAvailable:
                        self.__logger.warning('Got a task, but Worker is not available. Most probably is stopping right now')
                        writer.write(bytes([TaskScheduleStatus.FAILED.value]))
                    except Exception as e:
                        self.__logger.exception('no, cuz %s', e)
                        writer.write(bytes([TaskScheduleStatus.FAILED.value]))
                #
                # quit worker
                elif command == b'quit':
                    self.__worker.stop()
                    writer.write(b'\1')
                #
                # command drop/cancel current task
                elif command == b'drop':  # scheduler wants us to drop current task
                    try:
                        await self.__worker.cancel_task()
                    except:
                        self.__logger.exception('task drop failed')
                    writer.write(b'\1')
                #
                # command check worker status
                elif command == b'status':  # scheduler wants to know the status of current task
                    raise NotImplementedError()
                #
                # command to get worker's logs
                elif command == b'log':
                    invocation_id = struct.unpack('>Q', await reader.readexactly(8))[0]
                    for logfilepath in (self.__worker.get_log_filepath('output', invocation_id),
                                        self.__worker.get_log_filepath('error', invocation_id)):
                        if not os.path.exists(logfilepath):
                            writer.write(struct.pack('>I', 0))
                            continue
                        async with aiofiles.open(logfilepath, 'r') as f:
                            all_data = (await f.read()).encode('UTF-8')
                            writer.write(struct.pack('>I', len(all_data)))
                            writer.write(all_data)
                #
                # command get worker's log line by line
                elif command == b'logiter':  # scheduler wants to have the log:
                    raise NotImplementedError()
                    # TODO: redo this using line feeder from nethelpers! also pass invocation id
                    # logname = await asyncio.wait_for(reader.readline(), timeout=self.__timeout)
                    # if logname.endswith(b'\n'):
                    #     logname = logname[:-1]
                    # logname = logname.decode('UTF-8')
                    # logfilepath = self.__worker.get_log_filepath(logname)
                    # async with aiofiles.open(logfilepath, 'r') as f:
                    #     async for line in f:
                    #         writer.write(b'\x01')
                    #         writer.write(line)
                    #     writer.write(b'\x00')

                # now wait for answer to be delivered
                await writer.drain()
            else:
                raise NotImplementedError()
        except asyncio.TimeoutError:
            self.__logger.error('operation timeout')
            raise
        except EOFError:
            self.__logger.error('connection was abruptly closed')
            raise
        finally:
            writer.close()
            await writer.wait_closed()
            # according to the note in the beginning of the function - now reference can be cleared
            self.__saved_references.remove(asyncio.current_task())


class WorkerTaskClient:
    def __init__(self, ip: str, port: int, timeout=300.0):
        self.__logger = logging.get_logger('scheduler.workerconnection')
        self.__conn_task = asyncio.create_task(asyncio.open_connection(ip, port))
        self.__reader = None  # type: Optional[asyncio.StreamReader]
        self.__writer = None  # type: Optional[asyncio.StreamWriter]
        self.__timeout = timeout

    async def __aenter__(self) -> "WorkerTaskClient":
        await self._ensure_conn_open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        self.__writer.close()
        await self.__writer.wait_closed()

    async def _ensure_conn_open(self):
        if self.__reader is not None:
            return
        self.__reader, self.__writer = await self.__conn_task
        self.__writer.write(b'\0\0\0\0')

    async def ping(self):
        await self._ensure_conn_open()
        self.__writer.write(b'ping\n')
        try:
            await asyncio.wait_for(self.__writer.drain(), self.__timeout)
            pstatus, pvalue = struct.unpack('>Bl', await asyncio.wait_for(self.__reader.readexactly(5), self.__timeout))
            return WorkerPingReply(pstatus), pvalue
        except asyncio.exceptions.TimeoutError:
            self.__logger.error('network error')
            raise

    async def give_task(self, task: invocationjob.InvocationJob, reply_address):
        await self._ensure_conn_open()
        self.__writer.writelines([b'task\n'])
        taskserialized = await task.serialize_async()
        self.__writer.write(struct.pack('>I', len(taskserialized)))
        self.__writer.write(taskserialized)
        # now send where to respond
        # send URI-style link, dont limit urselves to ip
        if isinstance(reply_address, str):
            reply_address = reply_address.encode('UTF-8')
        self.__writer.write(struct.pack('>I', len(reply_address)))
        self.__writer.write(reply_address)
        await asyncio.wait_for(self.__writer.drain(), self.__timeout)
        reply = TaskScheduleStatus(ord(await asyncio.wait_for(self.__reader.readexactly(1), self.__timeout)))
        return reply

    async def cancel_task(self):
        await self._ensure_conn_open()
        self.__writer.write(b'drop\n')
        await asyncio.wait_for(self.__writer.drain(), self.__timeout)
        assert await self.__reader.readexactly(1) == b'\1'

    async def status(self):
        raise NotImplementedError()

    async def get_log(self, invocation_id) -> (str, str):
        self.__writer.write(b'log\n')
        self.__writer.write(struct.pack('>Q', invocation_id))
        await asyncio.wait_for(self.__writer.drain(), self.__timeout)
        stdoutsize = struct.unpack('>I', await asyncio.wait_for(self.__reader.readexactly(4), self.__timeout))[0]
        stdout = await asyncio.wait_for(self.__reader.readexactly(stdoutsize), self.__timeout)
        stderrsize = struct.unpack('>I', await asyncio.wait_for(self.__reader.readexactly(4), self.__timeout))[0]
        stderr = await asyncio.wait_for(self.__reader.readexactly(stderrsize), self.__timeout)
        return stdout.decode('UTF-8'), stderr.decode('UTF-8')

    async def get_log_iterate(self, logname: str, since=None):  # TODO: type it
        """
        get execution log from the worker
        :param logname: standard lognames are 'out' 'error' and shit like that
        :param since: since when to get logs. if None - get all the log
        :return:  iterable object to read log lines from
        """
        raise NotImplementedError()
        self.__writer.write(b'logiter\n')
        self.__writer.write(logname.encode('UTF-8'))
        await asyncio.wait_for(self.__writer.drain(), self.__timeout)
        return nethelpers.LineDrainer(self.__reader)



from . import enums
sql_init_script = '''
BEGIN TRANSACTION;
CREATE TABLE IF NOT EXISTS "lifeblood_metadata" (
	"version"	INTEGER NOT NULL DEFAULT 1,
	"component"	TEXT NOT NULL DEFAULT 'scheduler',
	"unique_db_id"	INTEGER DEFAULT NULL
);
CREATE TABLE IF NOT EXISTS "worker_groups" (
	"worker_hwid"	INTEGER NOT NULL,
	"group"	TEXT NOT NULL,
	FOREIGN KEY("worker_hwid") REFERENCES "resources"("hwid") ON UPDATE CASCADE ON DELETE CASCADE
);
CREATE TABLE IF NOT EXISTS "tasks" (
	"id"	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	"parent_id"	INTEGER,
	"children_count"	INTEGER NOT NULL DEFAULT 0,
	"active_children_count"	INTEGER NOT NULL DEFAULT 0,
	"state"	INTEGER NOT NULL,
	"state_details"	TEXT,
	"paused"	INTEGER DEFAULT 0,
	"node_id"	INTEGER NOT NULL,
	"node_input_name"	TEXT,
	"node_output_name"	TEXT,
	"work_data"	BLOB,
	"work_data_invocation_attempt"	INTEGER NOT NULL DEFAULT 0,
	"name"	TEXT,
	"attributes"	TEXT NOT NULL DEFAULT '{{}}',
	"split_level"	INTEGER NOT NULL DEFAULT 0,
	"priority"	REAL NOT NULL DEFAULT 50,
	"_invoc_requirement_clause"	TEXT,
	"environment_resolver_data"	BLOB,
	"dead"	INTEGER NOT NULL DEFAULT 0,
	FOREIGN KEY("node_id") REFERENCES "nodes"("id") ON UPDATE CASCADE ON DELETE RESTRICT,
	FOREIGN KEY("parent_id") REFERENCES "tasks"("id") ON UPDATE CASCADE ON DELETE SET NULL
);
CREATE TABLE IF NOT EXISTS "resources" (
	"hwid"	TEXT NOT NULL UNIQUE,
	"cpu_count"	INTEGER NOT NULL,
	"total_cpu_count"	INTEGER NOT NULL,
	"cpu_mem"	INTEGER NOT NULL,
	"total_cpu_mem"	INTEGER NOT NULL,
	"gpu_count"	INTEGER NOT NULL,
	"total_gpu_count"	INTEGER NOT NULL,
	"gpu_mem"	INTEGER NOT NULL,
	"total_gpu_mem"	INTEGER NOT NULL,
	PRIMARY KEY("hwid")
) WITHOUT ROWID;
CREATE TABLE IF NOT EXISTS "workers" (
	"id"	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	"last_address"	TEXT UNIQUE,
	"last_seen"	INTEGER,
	"last_checked"	INTEGER,
	"ping_state"	INTEGER NOT NULL,
	"state"	INTEGER NOT NULL,
	"worker_type"	INTEGER NOT NULL DEFAULT 0,
	"hwid"	INTEGER NOT NULL,
	FOREIGN KEY("hwid") REFERENCES "resources"("hwid") ON UPDATE CASCADE ON DELETE CASCADE
);
CREATE TABLE IF NOT EXISTS "task_splits" (
	"split_id"	INTEGER NOT NULL,
	"task_id"	INTEGER NOT NULL,
	"split_element"	INTEGER NOT NULL DEFAULT 0,
	"split_count"	INTEGER NOT NULL,
	"origin_task_id"	INTEGER NOT NULL,
	"split_sealed"	INTEGER NOT NULL DEFAULT 0,
	FOREIGN KEY("origin_task_id") REFERENCES "tasks"("id") ON UPDATE CASCADE ON DELETE RESTRICT,
	FOREIGN KEY("task_id") REFERENCES "tasks"("id") ON UPDATE CASCADE ON DELETE RESTRICT
);
CREATE TABLE IF NOT EXISTS "node_connections" (
	"id"	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	"node_id_in"	INTEGER NOT NULL,
	"node_id_out"	INTEGER NOT NULL,
	"in_name"	TEXT NOT NULL DEFAULT 'main',
	"out_name"	TEXT NOT NULL DEFAULT 'main',
	FOREIGN KEY("node_id_out") REFERENCES "nodes"("id") ON UPDATE CASCADE ON DELETE CASCADE,
	FOREIGN KEY("node_id_in") REFERENCES "nodes"("id") ON UPDATE CASCADE ON DELETE CASCADE,
	UNIQUE(node_id_in, node_id_out, in_name, out_name)
);
CREATE TABLE IF NOT EXISTS "nodes" (
	"id"	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	"type"	TEXT NOT NULL,
	"name"	TEXT,
	"node_object"	BLOB
);
CREATE TABLE IF NOT EXISTS "task_groups" (
	"task_id"	INTEGER NOT NULL,
	"group"	TEXT NOT NULL,
	FOREIGN KEY("task_id") REFERENCES "tasks"("id") ON UPDATE CASCADE ON DELETE CASCADE
);
CREATE TABLE IF NOT EXISTS "invocations" (
	"id"	INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	"task_id"	INTEGER NOT NULL,
	"worker_id"	NUMERIC NOT NULL,
	"node_id"	INTEGER NOT NULL,
	"state"	INTEGER NOT NULL,
	"return_code"	INTEGER,
	"log_external"	INTEGER NOT NULL DEFAULT 0,
	"stdout"	TEXT,
	"stderr"	TEXT,
	"progress"	REAL,
	"runtime"	REAL,
	FOREIGN KEY("worker_id") REFERENCES "workers"("id") ON UPDATE CASCADE ON DELETE RESTRICT,
	FOREIGN KEY("task_id") REFERENCES "tasks"("id") ON UPDATE CASCADE ON DELETE RESTRICT,
	FOREIGN KEY("node_id") REFERENCES "nodes"("id") ON UPDATE CASCADE ON DELETE RESTRICT
);
CREATE TABLE IF NOT EXISTS "task_group_attributes" (
	"group"	TEXT NOT NULL UNIQUE,
	"ctime"	INTEGER NOT NULL,
	"state"	INTEGER NOT NULL DEFAULT 0,
	"creator"	TEXT,
	"priority"	REAL NOT NULL DEFAULT 50,
	PRIMARY KEY("group")
);
CREATE INDEX IF NOT EXISTS "task_group_attrs_state_creator_idx" ON "task_group_attributes" (
	"status",
	"creator"
);
CREATE INDEX IF NOT EXISTS "task_dead" ON "tasks" (
	"dead"
);
CREATE INDEX IF NOT EXISTS "task_parent_id" ON "tasks" (
	"parent_id"
);
CREATE INDEX IF NOT EXISTS "invocation_worker_id_state" ON "invocations" (
	"worker_id",
	"state"
);
CREATE INDEX IF NOT EXISTS "task_groups_groups" ON "task_groups" (
	"group"
);
CREATE INDEX IF NOT EXISTS "invocation_worker_id" ON "invocations" (
	"worker_id"
);
CREATE INDEX IF NOT EXISTS "invocations_task_id" ON "invocations" (
	"task_id"
);
CREATE INDEX IF NOT EXISTS "task_groups_task_id" ON "task_groups" (
	"task_id"
);
CREATE INDEX IF NOT EXISTS "task_node_id" ON "tasks" (
	"node_id"
);
CREATE INDEX IF NOT EXISTS "task_splits_task_id" ON "task_splits" (
	"task_id"
);
CREATE INDEX IF NOT EXISTS "task_state" ON "tasks" (
	"state"
);
CREATE INDEX IF NOT EXISTS "task_state_paused_dead_idx" ON "tasks" (
	"state",
	"paused",
	"dead"
);
CREATE INDEX IF NOT EXISTS "task_dead_paused_state_idx" ON "tasks" (
	"dead",
	"paused",
	"state"
);
CREATE INDEX IF NOT EXISTS "invoc_state_idx" ON "invocations" (
	"state"
);
CREATE TRIGGER IF NOT EXISTS children_inserted
AFTER INSERT ON "tasks" WHEN new.parent_id IS NOT NULL
BEGIN
UPDATE "tasks" SET "children_count" = "children_count" + 1 WHERE "id" == new.parent_id;
END;
CREATE TRIGGER IF NOT EXISTS active_children_inserted
AFTER INSERT ON "tasks" WHEN new.state != {dead_state} AND new.parent_id IS NOT NULL
BEGIN
UPDATE "tasks" SET "active_children_count" = "active_children_count" + 1 WHERE "id" == new.parent_id;
END;
CREATE TRIGGER IF NOT EXISTS tasks_turning_dead
AFTER UPDATE OF "state" ON "tasks" WHEN old.state != {dead_state} AND new.state == {dead_state}
BEGIN
UPDATE "tasks" SET "active_children_count" = "active_children_count" - 1 WHERE "id" == new.parent_id;
UPDATE "tasks" SET "dead" = "dead" | 1 WHERE "id" == new."id";
END;
CREATE TRIGGER IF NOT EXISTS tasks_turning_undead
AFTER UPDATE OF "state" ON "tasks" WHEN old.state == {dead_state} AND new.state != {dead_state}
BEGIN
UPDATE "tasks" SET "active_children_count" = "active_children_count" + 1 WHERE "id" == new.parent_id;
UPDATE "tasks" SET "dead" = "dead" & ~1 WHERE "id" == new."id";
END;
CREATE TRIGGER IF NOT EXISTS task_group_turning_unarchived
AFTER UPDATE OF "state" ON "task_group_attributes" WHEN old.state == 1 AND new.state != 1
BEGIN
UPDATE "tasks" SET "dead" = "dead" & ~2 WHERE "id" IN (
	SELECT "task_id" FROM task_groups
	WHERE "group" == new."group"
);
END;
CREATE TRIGGER IF NOT EXISTS task_group_turning_archived
AFTER UPDATE OF "state" ON "task_group_attributes" WHEN old.state != 1 AND new.state == 1
BEGIN
UPDATE "tasks" SET "dead" = "dead" | 2 WHERE "id" IN (
	SELECT "task_id" FROM task_groups
	WHERE "group" == new."group"
);
END;
CREATE TRIGGER IF NOT EXISTS flush_task_state
BEFORE UPDATE OF "state" ON "tasks" WHEN old.state <> new.state
BEGIN
UPDATE "tasks" SET "state_details" = NULL WHERE "id" == new.id;
END;
CREATE TRIGGER IF NOT EXISTS flush_task_input_output_names
BEFORE UPDATE OF "node_id" ON "tasks" WHEN old.node_id <> new.node_id
BEGIN
UPDATE "tasks" SET "node_output_name" = NULL WHERE "id" == new.id;
END;
COMMIT;
PRAGMA journal_mode=wal;
PRAGMA synchronous=NORMAL;
'''.format(dead_state=enums.TaskState.DEAD.value)
# PRAGMA soft_heap_limit=100000000;
# PRAGMA mmap_size=100000000;
# TODO: add after delete triggers for children count


worker_resource_db_init_script = '''
BEGIN TRANSACTION;
CREATE TABLE IF NOT EXISTS "resources" (
	"pid"	INTEGER NOT NULL,
	"cpu_count"	INTEGER NOT NULL,
	"cpu_mem"	INTEGER NOT NULL,
	"gpu_count"	INTEGER NOT NULL,
	"gpu_mem"	INTEGER NOT NULL,
	PRIMARY KEY("pid")
);
COMMIT;
PRAGMA journal_mode=wal;
PRAGMA synchronous=NORMAL;
'''


import sys
import subprocess
from pathlib import Path


def main():
    path_to_stuff = Path(__package__)/'stock_nodes'/'matrixclient'/'data'/'matrixclient.pyz'
    sys.exit(subprocess.Popen([sys.executable, path_to_stuff, *sys.argv[1:]]).wait())


if __name__ == '__main__':
    sys.exit(main() or 0)



from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionTexturing


class AlicevisionTexturing(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionTexturing, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('input', 'Dense SfM Data', NodeParameterType.STRING, "`task['av_dense_point_cloud']`")
            ui.add_parameter('imagesFolder', 'Images Folder', NodeParameterType.STRING, "`task['av_prepare_dense']`")
            ui.add_parameter('inputMesh', 'Input Mesh', NodeParameterType.STRING, "`task['av_mesh_object_filtered']`")
            ui.add_separator()
            ui.add_parameter('output', 'Output Dense Point Cloud', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/Texturing")
            ui.add_separator()
            ui.add_parameter('textureSide', 'Texture Side', NodeParameterType.INT, 8192)
            ui.add_parameter('downscale', 'Texture Downscale', NodeParameterType.INT, 2)
            ui.add_parameter('outputTextureFileType', 'Texture File Type', NodeParameterType.STRING, 'png')\
                .add_menu((('png', 'png'), ('jpg', 'jpg'), ('tiff', 'tiff'), ('exr', 'exr')))
            ui.add_parameter('unwrapMethod', 'Unwrap Method', NodeParameterType.STRING, 'Basic')\
                .add_menu((('Basic', 'Basic'), ('LSCM', 'LSCM'), ('ABF', 'ABF')))
            ui.add_parameter('useUDIM', 'Use UDIM', NodeParameterType.BOOL, True)
            ui.add_parameter('fillHoles', 'Fill Holes', NodeParameterType.BOOL, False)
            ui.add_parameter('correctEV', 'Correct Exposure', NodeParameterType.BOOL, False)

    @classmethod
    def label(cls) -> str:
        return 'AV Texturing'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'texturing', 'mesh', 'texture'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_texturing'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output = Path(context.param_value('output').strip())
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.mkdir(parents=True, exist_ok=True)

        args = ['aliceVision_texturing', '--verboseLevel', 'info']
        args += ['--input', context.param_value('input').strip()]
        args += ['--imagesFolder', context.param_value('imagesFolder').strip()]
        args += ['--inputMesh', context.param_value('inputMesh').strip()]
        for param_name in ('textureSide', 'downscale', 'outputTextureFileType',
                           'unwrapMethod', 'useUDIM', 'fillHoles', 'correctEV'):
            args += [f'--{param_name}', context.param_value(param_name)]

        args += ['--output', output]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_output_textured_folder', str(output))
        res.set_attribute('av_output_textured_obj', str(output/'texturedMesh.obj'))
        res.set_attribute('av_output_textured_mtl', str(output/'texturedMesh.mtl'))
        return res



import time

from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.processingcontext import ProcessingContext
from lifeblood.nodethings import ProcessingResult
from lifeblood.uidata import NodeParameterType

from types import MappingProxyType
from typing import TYPE_CHECKING, Iterable
if TYPE_CHECKING:
    from lifeblood.scheduler import Scheduler


def node_class():
    return Python


process_help = "# use task variable to get/set attribs, e.g.     task['attr'] = 123\n" \
               "# use schedule() to decide if you want to create an invocation to be done by worker\n" \
               "# this code will be executed on scheduler, so don't make in heavy.\n" \
               "# leave heavy things for 'invoke' parameter below\n" \
               "\n" \
               "# schedule()  # uncomment this to launch code in invoke parameter below on a remote worker\n" \
               "\n"

invoke_help = "# use task variable to get attribs, e.g.     task['attr'] = 123\n" \
              "# this code will be executed on remote worker, all calls to task['smth'] will be resolved beforehand\n" \
              "\n"


class Python(BaseNodeWithTaskRequirements):
    def __init__(self, name: str):
        super(Python, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.4, 0.4, 0.2)
            ui.add_parameter('process', 'process', NodeParameterType.STRING, process_help).set_text_multiline('python')
            ui.add_parameter('invoke', 'invoke', NodeParameterType.STRING, invoke_help).set_text_multiline('python')
            # ui.add_parameter('postprocess', 'postprocess', NodeParameterType.STRING, '').set_text_multiline('python')

    @classmethod
    def label(cls) -> str:
        return 'python script'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'script', 'python', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'python'

    def process_task(self, context) -> ProcessingResult:
        class _TaskHelper:
            def __init__(self, context: ProcessingContext, readonly: bool):
                self.__context = context
                self.__overrides = {}
                self.__todel = set()
                self.__readonly = readonly

            def __getitem__(self, item):
                if item in self.__todel:
                    raise KeyError(f'key {item} was deleted')
                if item in self.__overrides:
                    return self.__overrides[item]
                return self.__context.task_attribute(item)

            def __setitem__(self, key, value):
                if self.__readonly:
                    raise RuntimeError('cannot set attrib for readonly task')
                if key in self.__todel:
                    self.__todel.remove(key)
                self.__overrides[key] = value

            def __delitem__(self, key):
                if key in self.__todel:
                    raise KeyError(f'key {key} was already deleted')
                if key in self.__overrides:
                    del self.__overrides[key]
                self.__todel.add(key)

            def __getattr__(self, item):
                if not self.__context.task_has_field(item):
                    raise AttributeError(f'task has not field {item}')
                return self.__context.task_field(item)

            def attribute_names(self):
                return set(self.__context.task_attributes().keys()) | set(self.__overrides.keys())

            def _overrides(self):
                return self.__overrides

            def _removes(self):
                return self.__todel

        def _set_sched():
            nonlocal do_schedule
            do_schedule = True

        task_helper = _TaskHelper(context, readonly=False)

        exec_locals = {'task': task_helper, 'schedule': _set_sched}
        do_schedule = False
        exec(context.param_value('process'), exec_locals)

        if do_schedule:
            inv = InvocationJob(['python', ':/main_invocation.py'])
            inok_task_attributes = {k: task_helper[k] for k in task_helper.attribute_names()}
            # TODO: task here is not the same! no __getattr__ !
            inv.set_extra_file('main_invocation.py', f'task = {repr(inok_task_attributes)}\n{context.param_value("invoke")}')
        else:
            inv = None
        res = ProcessingResult(inv)

        for k, v in task_helper._overrides().items():
            res.set_attribute(k, v)

        for k in task_helper._removes():
            res.remove_attribute(k)

        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import sys
import traceback
import json
import itertools
import threading  # for bugfix
from concurrent.futures import ThreadPoolExecutor
import aiosqlite
import asyncio
import time
from .. import logging
from ..enums import WorkerState, InvocationState, TaskState, TaskGroupArchivedState, TaskScheduleStatus
from ..misc import atimeit
from ..worker_messsage_processor import WorkerControlClient
from ..invocationjob import InvocationJob
from ..environment_resolver import EnvironmentResolverArguments
from ..nodethings import ProcessingResult
from ..exceptions import *
from .. import pluginloader
from .. import aiosqlite_overlay
from ..config import get_config
from ..ui_events import TaskData, TaskDelta
from ..net_messages.address import AddressChain

from .scheduler_component_base import SchedulerComponentBase

from typing import Any, List, Optional, TYPE_CHECKING

if TYPE_CHECKING:  # TODO: maybe separate a subset of scheduler's methods to smth like SchedulerData class, or idunno, for now no obvious way to separate, so having a reference back
    from .scheduler import Scheduler


# import tracemalloc
# tracemalloc.start()

class TaskProcessor(SchedulerComponentBase):
    def __init__(self, scheduler: "Scheduler"):
        super().__init__(scheduler)
        self.__logger = logging.get_logger('scheduler.task_processor')

        self.__processing_interval = 5  # we don't need interval too small as now things may kick processor out of sleep as needed
        self.__processing_interval_mult = 1
        self.__dormant_mode_processing_interval_multiplier = 5

        self.awaiter_lock = asyncio.Lock()
        # task processing coroutimes
        self.awaiter_executor = ThreadPoolExecutor(thread_name_prefix='awaiter')  # TODO: max_workers= set from config

        config = get_config('scheduler')
        self.__invocation_attempts = config.get_option_noasync('invocation.default_attempts', 3)  # TODO: config should be directly used when needed to allow dynamically reconfigure running scheduler

    def _main_task(self):
        return self.task_processor()

    def _my_wake(self):
        self.__logger.info('exiting DORMANT mode. mode is STANDARD now')
        self.__processing_interval_mult = 1
        self.poke()

    def _my_sleep(self):
        self.__logger.info('entering DORMANT mode')
        self.__processing_interval_mult = self.__dormant_mode_processing_interval_multiplier

    @atimeit()
    async def _awaiter(self, processor_to_run, task_row, abort_state: TaskState, skip_state: TaskState):  # TODO: process task generation errors
        _bench_point_0 = time.perf_counter()
        task_id = task_row['id']
        loop = asyncio.get_event_loop()

        try:
            async with self.scheduler.get_node_lock_by_id(task_row['node_id']).reader_lock:
                time_processing_start = time.perf_counter()
                process_result: ProcessingResult = await loop.run_in_executor(self.awaiter_executor, processor_to_run, task_row)  # TODO: this should have task and node attributes!
                self.__logger.debug(f'(post)processing for {task_row["node_id"]} took {time.perf_counter()-time_processing_start:.4f}')
        except NodeNotReadyToProcess:
            async with self.awaiter_lock, self.scheduler.data_access.lazy_data_transaction('awaiter_con') as con:
                await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                                  (abort_state.value, task_id))
                con.add_after_commit_callback(self.scheduler.ui_state_access.scheduler_reports_task_updated, TaskDelta(task_id, state=abort_state))
                await con.commit(self.poke)
            self.__logger.debug('node reports: not ready to process yet')
            return
        except Exception as e:
            async with self.awaiter_lock, self.scheduler.data_access.lazy_data_transaction('awaiter_con') as con:
                state_details = json.dumps({'message': traceback.format_exc(),
                                            'happened_at': task_row['state'],
                                            'type': 'exception',
                                            'exception_str': str(e),
                                            'exception_type': str(type(e))})
                await con.execute('UPDATE tasks SET "state" = ?, "state_details" = ? WHERE "id" = ?',
                                  (TaskState.ERROR.value,
                                   state_details,
                                   task_id))
                con.add_after_commit_callback(self.scheduler.ui_state_access.scheduler_reports_task_updated, TaskDelta(task_id, state=TaskState.ERROR, state_details=state_details))
                await con.commit(self.poke)
                self.__logger.exception('error happened %s %s', type(e), e)
            return

        _bench_point_1 = time.perf_counter()
        # why is there lock? it looks locking manually is waaaay more efficient than relying on transaction locking
        async with self.awaiter_lock, self.scheduler.data_access.lazy_data_transaction('awaiter_con') as con:
            # con.row_factory = aiosqlite.Row
            # This implicitly starts transaction

            _bench_point_2 = time.perf_counter()
            ui_task_delta = TaskDelta(task_id)  # for ui event
            ui_task_delta_split = None

            if not con.in_transaction:
                await con.execute('BEGIN IMMEDIATE')
                assert con.in_transaction
            _bench_point_3 = time.perf_counter()
            if process_result.output_name:
                await con.execute('UPDATE tasks SET "node_output_name" = ? WHERE "id" = ?',
                                  (process_result.output_name, task_id))
                ui_task_delta.node_output_name = process_result.output_name  # for ui event
            _bench_point_4 = time.perf_counter()

            # note: this may be not obvious, but ALL branches of the next if result in implicit transaction start
            if process_result.do_kill_task:
                await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                                  (TaskState.DEAD.value, task_id))
                ui_task_delta.state = TaskState.DEAD  # for ui event
            else:
                if process_result.invocation_job is None:  # if no job to do
                    await con.execute('UPDATE tasks SET "work_data" = ?, "work_data_invocation_attempt" = 0, "state" = ?, "_invoc_requirement_clause" = ? '
                                      'WHERE "id" = ?',
                                      (None, skip_state.value, None,
                                       task_id))
                    ui_task_delta.work_data_invocation_attempt = 0  # for ui event
                    ui_task_delta.state = skip_state  # for ui event
                else:
                    # if there is an invocation - we force environment wrapper arguments from task onto it
                    if task_row['environment_resolver_data'] is not None:
                        process_result.invocation_job._set_envresolver_arguments(await EnvironmentResolverArguments.deserialize_async(task_row['environment_resolver_data']))

                    taskdada_serialized = await process_result.invocation_job.serialize_async()
                    invoc_requirements_sql = process_result.invocation_job.requirements().final_where_clause()
                    invoc_requirements_dict_str = json.dumps(process_result.invocation_job.requirements().to_dict(resources_only=True))
                    job_priority = process_result.invocation_job.priority()
                    async with con.execute('SELECT MAX(task_group_attributes.priority) AS priority FROM task_group_attributes '
                                           'INNER JOIN task_groups ON task_group_attributes."group"==task_groups."group" '
                                           'WHERE task_groups.task_id==? AND task_group_attributes.state==?', (task_id, TaskGroupArchivedState.NOT_ARCHIVED.value)) as cur:
                        group_priority = await cur.fetchone()
                        if group_priority is None:
                            group_priority = 50.0  # "or" should only work in case there were no unarchived groups at all for the task
                        else:
                            group_priority = group_priority[0] or 50.0
                    await con.execute('UPDATE tasks SET "work_data" = ?, "work_data_invocation_attempt" = 0, "state" = ?, "_invoc_requirement_clause" = ?, '
                                      'priority = ? '
                                      'WHERE "id" = ?',
                                      (taskdada_serialized, TaskState.READY.value, ':::'.join((invoc_requirements_sql, invoc_requirements_dict_str)),
                                       group_priority + job_priority,
                                       task_id))
                    ui_task_delta.work_data_invocation_attempt = 0  # for ui event
                    ui_task_delta.state = TaskState.READY  # for ui event

            _bench_point_5 = time.perf_counter()
            if process_result.do_split_remove:  # TODO: check that there is no race conditions among splitted tasks to remove the split
                async with con.execute('SELECT split_sealed FROM task_splits WHERE split_id = ?', (task_row['split_id'],)) as sealcur:
                    res = await sealcur.fetchone()
                if res is not None and res['split_sealed'] == 0:  # sealing split does actually exist and not sealed
                    # async with con.execute('SELECT task_id FROM task_splits WHERE split_id = ?', (task_row['split_id'])) as tcur:
                    #     task_ids_to_update = [x['task_id'] for x in await tcur.fetchall()]
                    # await con.executemany('UPDATE tasks set "state" = ? WHERE "id" = ?', ((TaskState.DEAD.value, x) for x in task_ids_to_update))
                    await con.execute('UPDATE task_splits SET "split_sealed" = 1 '
                                      'WHERE  "split_id" = ?',
                                      (task_row['split_id'],))
                    # teleport original task to us
                    await con.execute('UPDATE tasks SET "node_id" = ?, "state" = ? WHERE "id" = ?',
                                      (task_row['node_id'], TaskState.DONE.value, task_row['split_origin_task_id']))
                    ui_task_delta_split = TaskDelta(task_row['split_origin_task_id'], node_id=task_row['node_id'], state=TaskState.DONE)  # for ui event
                    if process_result.output_name:
                        await con.execute('UPDATE tasks SET "node_output_name" = ? WHERE "id" = ?',
                                          (process_result.output_name, task_row['split_origin_task_id']))
                        ui_task_delta_split.node_output_name = process_result.output_name  # for ui event
                        # so sealed split task will get the same output_name as the task that is sealing the split
                    # and update its attributes if provided
                    if len(process_result.split_attributes_to_set) > 0:
                        async with con.execute('SELECT attributes FROM tasks WHERE "id" = ?', (task_row['split_origin_task_id'],)) as attcur:
                            attributes = await asyncio.get_event_loop().run_in_executor(None, json.loads, (await attcur.fetchone())['attributes'])
                            attributes.update(process_result.split_attributes_to_set)
                            for k, v in process_result.split_attributes_to_set.items():
                                if v is None:
                                    del attributes[k]
                            result_serialized = await asyncio.get_event_loop().run_in_executor(None, json.dumps, attributes)
                            await con.execute('UPDATE tasks SET "attributes" = ? WHERE "id" = ?',
                                              (result_serialized, task_row['split_origin_task_id']))

            _bench_point_6 = time.perf_counter()
            _bench_point_7 = _bench_point_6
            if process_result.attributes_to_set:  # not None or {}
                attributes = await asyncio.get_event_loop().run_in_executor(None, json.loads, task_row['attributes'] or '{}')
                attributes.update(process_result.attributes_to_set)
                for k, v in process_result.attributes_to_set.items():  # TODO: hmmm, None is a valid value...
                    if v is None:
                        del attributes[k]
                result_serialized = await asyncio.get_event_loop().run_in_executor(None, json.dumps, attributes)
                _bench_point_7 = time.perf_counter()
                await con.execute('UPDATE tasks SET "attributes" = ? WHERE "id" = ?',
                                  (result_serialized, task_id))
            _bench_point_8 = time.perf_counter()

            # process environment resolver arguments if provided
            if (envargs := process_result._environment_resolver_arguments) is not None:
                await con.execute('UPDATE tasks SET environment_resolver_data = ? WHERE "id" = ?',
                                  (await envargs.serialize_async(), task_id))

            _bench_point_9 = time.perf_counter()
            # spawning new tasks after all attributes were set, so children inherit
            # spawn
            if process_result.spawn_list is not None:
                for spawn in process_result.spawn_list:
                    # we do NOT allow spawning children anywhere else but in the same node, and with the task as parent
                    spawn.force_set_node_task_id(task_row['node_id'], task_row['id'])
                await self.scheduler.spawn_tasks(process_result.spawn_list, con=con)

            _bench_point_10 = time.perf_counter()
            # splits
            if process_result._split_attribs is not None:
                split_count = len(process_result._split_attribs)
                for attr_dict, split_task_id in zip(process_result._split_attribs, await self.split_task(task_id, split_count, con)):
                    async with con.execute('SELECT attributes FROM "tasks" WHERE "id" = ?', (split_task_id,)) as cur:
                        split_task_dict = await cur.fetchone()
                    assert split_task_dict is not None
                    split_task_attrs = await asyncio.get_event_loop().run_in_executor(None, json.loads, split_task_dict['attributes'])
                    split_task_attrs.update(attr_dict)
                    await con.execute('UPDATE "tasks" SET attributes = ? WHERE "id" = ?', (json.dumps(split_task_attrs), split_task_id))  # TODO: run dumps in executor

            _bench_point_11 = time.perf_counter()
            con.add_after_commit_callback(self.scheduler.ui_state_access.scheduler_reports_tasks_updated, [ui_task_delta] if ui_task_delta_split is None else [ui_task_delta, ui_task_delta_split])  # ui event
            await con.commit(self.poke)
            _bench_point_12 = time.perf_counter()
        self.__logger.debug(
            f'_awaiter all: {_bench_point_12 - _bench_point_0}: '
            f'process: {_bench_point_0 - _bench_point_1}, '
            f'lock: {_bench_point_1 - _bench_point_2}, '
            f'tran_begin: {_bench_point_3 - _bench_point_2},'
            f'upd_out: {_bench_point_4 - _bench_point_3}, '
            f'upd_rest: {_bench_point_5 - _bench_point_4},'
            f'split_rm: {_bench_point_6 - _bench_point_5}, '
            f'attr_set: {_bench_point_7 - _bench_point_6} + {_bench_point_8 - _bench_point_7}, '
            f'envr_set: {_bench_point_9 - _bench_point_8}, '
            f'spawn: {_bench_point_10 - _bench_point_9}, '
            f'split: {_bench_point_11 - _bench_point_10}, '
            f'commit: {_bench_point_12 - _bench_point_11}')

    # submitter
    @atimeit()
    async def _submitter(self, task_row, worker_row):
        self.__logger.debug(f'submitter started')
        try:
            addr = AddressChain(worker_row['last_address'])
        except ValueError:
            self.__logger.error('error address converting during unexpected here. ping should have cought it')
            addr = None  # set to invalid values to exit in error-checking if a bit below
            # TODO: add nicer error check for invalid address. currently we fail to open connection to None

        task_id = task_row['id']
        ui_task_delta = TaskDelta(task_id)  # for ui event
        work_data = task_row['work_data']
        assert work_data is not None
        task: InvocationJob = await asyncio.get_event_loop().run_in_executor(None, InvocationJob.deserialize, work_data)
        if not task.args():
            async with self.awaiter_lock, self.scheduler.data_access.data_connection() as skipwork_transaction:
                await skipwork_transaction.execute('UPDATE tasks SET state = ? WHERE "id" = ?',
                                                   (TaskState.POST_WAITING.value, task_id))
                await skipwork_transaction.execute('UPDATE workers SET state = ? WHERE "id" = ?',
                                                   (WorkerState.IDLE.value, worker_row['id']))
                # unset resource usage
                await self.scheduler._update_worker_resouce_usage(worker_row['id'], hwid=worker_row['hwid'], connection=skipwork_transaction)
                skipwork_transaction.add_after_commit_callback(self.scheduler.ui_state_access.scheduler_reports_task_updated, TaskDelta(task_id, state=TaskState.POST_WAITING))
                await skipwork_transaction.commit()
                return

        # so task.args() is not None
        async with self.scheduler.data_access.data_connection() as submit_transaction:
            submit_transaction.row_factory = aiosqlite.Row
            async with self.awaiter_lock:
                async with submit_transaction.execute(
                        'INSERT INTO invocations ("task_id", "worker_id", "state", "node_id") VALUES (?, ?, ?, ?)',
                        (task_id, worker_row['id'], InvocationState.INVOKING.value, task_row['node_id'])) as incur:
                    invocation_id = incur.lastrowid  # rowid should be an alias to id, acc to sqlite manual
                await submit_transaction.commit()

            task._set_invocation_id(invocation_id)
            task._set_task_id(task_id)
            async with submit_transaction.execute('SELECT attributes FROM tasks WHERE "id" == ?', (task_id,)) as attcur:
                task_attributes_raw = ((await attcur.fetchone()) or ['{}'])[0]
            task._set_task_attributes(await asyncio.get_event_loop().run_in_executor(None, json.loads, task_attributes_raw))
            self.__logger.debug(f'submitting task to {addr}')
            try:
                # this is potentially a long operation - db must NOT be locked during it
                with WorkerControlClient.get_worker_control_client(addr, self.scheduler.message_processor()) as client:  # type: WorkerControlClient
                    # import random
                    # await asyncio.sleep(random.uniform(0, 8))  # DEBUG! IMITATE HIGH LOAD
                    reply = await client.give_task(task, self.scheduler.server_message_address())
                self.__logger.debug(f'got reply {reply}')
            except Exception as e:
                self.__logger.error('some unexpected error %s %s' % (str(type(e)), str(e)))
                reply = TaskScheduleStatus.FAILED

            async with self.awaiter_lock:
                await submit_transaction.execute('BEGIN IMMEDIATE')
                async with submit_transaction.execute('SELECT "state" FROM workers WHERE "id" == ?', (worker_row['id'],)) as incur:
                    worker_state = WorkerState((await incur.fetchone())[0])
                async with submit_transaction.execute('SELECT "state" FROM invocations WHERE "id" == ?', (invocation_id,)) as incur:
                    # if worker managed to stop and start before we reach this transaction - invocation state will be reset
                    # we have to check it
                    maybe_updated_invocation_state = InvocationState((await incur.fetchone())[0])

                worker_apparently_restarted = False
                if maybe_updated_invocation_state != InvocationState.INVOKING:
                    self.__logger.warning(f'worker seem to have stopped during submission attempt, ignoring, retrying. reply was: {reply}, worker state is: {worker_state}')
                    worker_apparently_restarted = True
                    reply = TaskScheduleStatus.FAILED

                # IF worker state is NOT invoking - then either worker_hello, or worker_bye happened between starting _submitter and here
                if worker_state == WorkerState.OFF:
                    self.__logger.debug('submitter: worker state changed to OFF during submitter work')
                    if reply == TaskScheduleStatus.SUCCESS:
                        self.__logger.warning('submitter succeeded, yet worker state changed to OFF in the middle of submission. forcing reply to FAIL')
                        reply = TaskScheduleStatus.FAILED

                # this assert should never break: as hello preserves INVOKING state, and we catch worker restart case
                assert worker_apparently_restarted or worker_state != WorkerState.IDLE, f'worker restarted={worker_apparently_restarted}, state={worker_state}'

                if reply == TaskScheduleStatus.SUCCESS:
                    await submit_transaction.execute('UPDATE tasks SET state = ?, '
                                                     '"work_data_invocation_attempt" = "work_data_invocation_attempt" + 1 '
                                                     'WHERE "id" = ?',
                                                     (TaskState.IN_PROGRESS.value, task_id))
                    ui_task_delta.state = TaskState.IN_PROGRESS  # for ui event
                    async with submit_transaction.execute('SELECT "work_data_invocation_attempt" FROM tasks WHERE "id" == ?', (task_id,)) as tmpcur:  # TODO: remove this extra query!  # ui event
                        ui_task_delta.work_data_invocation_attempt = (await tmpcur.fetchone())['work_data_invocation_attempt']  # TODO: maybe remove work_data_invocation_attempt from generic ui events/data at all  # ui event
                    await submit_transaction.execute('UPDATE workers SET state = ? WHERE "id" = ?',
                                                     (WorkerState.BUSY.value, worker_row['id']))
                    await submit_transaction.execute('UPDATE invocations SET state = ? WHERE "id" = ?',
                                                     (InvocationState.IN_PROGRESS.value, invocation_id))
                else:  # on anything but success - cancel transaction
                    self.__logger.debug(f'submitter failed, rolling back for wid {worker_row["id"]}')
                    await submit_transaction.execute('UPDATE tasks SET state = ? WHERE "id" = ?',
                                                     (TaskState.READY.value,
                                                      task_id))
                    ui_task_delta.state = TaskState.READY  # for ui event
                    await submit_transaction.execute('UPDATE workers SET state = ? WHERE "id" = ?',
                                                     (WorkerState.IDLE.value if worker_state != WorkerState.OFF else WorkerState.OFF.value,
                                                      worker_row['id']))
                    await submit_transaction.execute('DELETE FROM invocations WHERE "id" = ?',
                                                     (invocation_id,))
                    # update resource usage to none
                    await self.scheduler._update_worker_resouce_usage(worker_row['id'], hwid=worker_row['hwid'], connection=submit_transaction)
                    # TODO: if task was reverted we MIGHT need to poke scheduler after transaction commit to process task again straight awau
                    #  Or something else, have other things to do now, think about this one later
                submit_transaction.add_after_commit_callback(self.scheduler.ui_state_access.scheduler_reports_task_updated, ui_task_delta)  # ui event
                await submit_transaction.commit()

    async def task_processor(self):
        # this will hold references to tasks created with asyncio.create_task
        tasks_to_wait = set()
        stop_task = asyncio.create_task(self._stop_event.wait())
        kick_wait_task = asyncio.create_task(self._poke_event.wait())
        gc_counter = 0
        # tm_counter = 0
        self._main_task_is_ready_now()
        while not self._stop_event.is_set():
            data_access = self.scheduler.data_access
            gc_counter += 1
            # tm_counter += 1
            if gc_counter >= 120:  # TODO: to config this timing
                gc_counter = 0
                self.__logger.debug('========')
                self.__logger.debug('================================================================')
                with threading._shutdown_locks_lock:
                    self.__logger.debug(f'loose threads: {len(threading._shutdown_locks)}')
                    threading._shutdown_locks.difference_update([lock for lock in threading._shutdown_locks if not lock.locked()])
                    self.__logger.debug(f'loose threads after cleanup: {len(threading._shutdown_locks)}')
                self.__logger.debug(f'total tasks: {len(asyncio.all_tasks())}')

                def _gszofdr(obj):
                    sz = sys.getsizeof(obj)
                    for k, v in obj.items():
                        sz += sys.getsizeof(k)
                        sz += sys.getsizeof(v)
                        if isinstance(v, dict):
                            sz += _gszofdr(v)
                    return sz

                # pruning db_cache
                async with data_access.data_connection() as con:
                    con.row_factory = aiosqlite.Row
                    async with con.execute('SELECT "id" FROM invocations WHERE state == ?',
                                           (InvocationState.IN_PROGRESS.value,)) as inv:
                        filtered_invocs = set(x['id'] for x in await inv.fetchall())
                for inv in tuple(data_access.mem_cache_invocations.keys()):
                    if inv not in filtered_invocs:  # Note: since task finish/cancel reporting is in the same thread as this - there will not be race conditions for del, as there's no await
                        del data_access.mem_cache_invocations[inv]
                filtered_invocs.clear()
                # prune done

                self.__logger.debug(f'size of temp db cache: {_gszofdr({1: data_access.mem_cache_invocations, 2: data_access.mem_cache_workers_resources, 3: data_access.mem_cache_workers_state})}')
                self.__logger.debug('================================================================')
                self.__logger.debug('========')

                # self.__logger.debug(f'\n\n {mem_top(verbose_types=[set], limit=16)} \n\n')
                #  seems that memtop's gc calls cause some random exceptions on db's fetch all
                #  https://bugs.python.org/issue37788
                #  https://bugs.python.org/issue15108
                #  also https://gist.github.com/ulope/db811b6cf853ff267f27e4295bc4739e
                # import gc
                # objs = gc.get_objects()
                # objs = sorted(objs, key=lambda obj: len(gc.get_referents(obj)), reverse=True)
                # print(repr(gc.get_referrers(objs[0]))[:200])
                # print('\n')
                # print(repr(gc.get_referrers(objs[1]))[:200])
            # if tm_counter >= 10*60*2:
            #     tm_counter = 0
            #     snapshot = tracemalloc.take_snapshot()
            #     top_stats = snapshot.statistics('lineno')
            #     self.__logger.warning('\n\n[ Top 10 MEM USERS]\n{}\n\n'.format("\n".join(str(stat) for stat in top_stats[:10])))

            # first prune awaited tasks
            to_remove = set()
            for task_to_wait in tasks_to_wait:
                if task_to_wait.done():
                    to_remove.add(task_to_wait)
                    try:
                        await task_to_wait
                    except Exception as e:
                        self.__logger.exception('awaited task raised some problems')
            tasks_to_wait -= to_remove

            # now proceed with processing
            _debug_con = time.perf_counter()
            total_processed = 0
            total_state_changes = 0  # note that total_state_changes may be greater than total_processed, as total_processed refers to existing tasks only, but total_state_changes counts new splits as well
            async with data_access.data_connection() as con:
                con.row_factory = aiosqlite.Row

                for task_state in (TaskState.WAITING, TaskState.READY, TaskState.DONE, TaskState.POST_WAITING, TaskState.SPAWNED):
                    _debug_sel = time.perf_counter()
                    async with con.execute('SELECT tasks.id, tasks.parent_id, tasks.children_count, tasks.active_children_count, tasks.state, {attrs}'
                                           'tasks.node_id, tasks.node_input_name, tasks.node_output_name, tasks.name, tasks.split_level, '
                                           'tasks.work_data, tasks.work_data_invocation_attempt, tasks._invoc_requirement_clause, '
                                           'nodes.type as node_type, nodes.id as node_id, '
                                           'task_splits.split_id as split_id, task_splits.split_element as split_element, task_splits.split_count as split_count, task_splits.origin_task_id as split_origin_task_id '
                                           'FROM tasks INNER JOIN nodes ON tasks.node_id=nodes.id '
                                           'LEFT JOIN task_splits ON tasks.id=task_splits.task_id '
                                           'WHERE (state = ?) '
                                           'AND paused = 0 '
                                           'AND dead = 0 '
                                           'ORDER BY {prio_sort} RANDOM()'.format(
                            prio_sort='tasks.priority DESC, ' if task_state == TaskState.READY else '',
                            attrs='attributes, environment_resolver_data, ' if task_state in (TaskState.WAITING, TaskState.POST_WAITING) else ''
                            ),

                                           (task_state.value,)) as cur:
                        all_task_rows = await cur.fetchall()  # we dont want to iterate reading over changing rows - easy to deadlock yourself (as already happened)
                        # if too much tasks here - consider adding LIMIT to execute and work on portions only

                    _debug_pstart = time.perf_counter()
                    if _debug_pstart - _debug_sel > 0.05:  # we shouldn't even worry if it's less
                        self.__logger.debug(f'SELECT took {_debug_pstart - _debug_sel}')

                    if len(all_task_rows) == 0:
                        continue
                    total_processed += len(all_task_rows)

                    self.__logger.debug(f'total {task_state.name}: {len(all_task_rows)}')
                    # TODO: the problem might occur below when there are thousands of processing tasks - it may take some time before implicit transaction lock is given to task_processor
                    #
                    # waiting to be processed
                    if task_state == TaskState.WAITING:
                        awaiters = []
                        set_to_stuff = []
                        for task_row in all_task_rows:
                            if task_row['node_type'] not in pluginloader.plugins:
                                self.__logger.error(f'plugin to process "{task_row["node_type"]}" not found!')
                                # await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                                #                   (TaskState.ERROR.value, task_row['id']))
                                set_to_stuff.append((TaskState.ERROR.value, task_row['id']))
                                total_state_changes += 1
                            else:
                                # note that ready_to_process_task is ran not inside the read lock
                                # as it's expected that:
                                #  - running the function is even faster than locking
                                #  - function misfire (being highly unlikely) does not have side effects, so will not cause any damage
                                try:
                                    if not (await self.scheduler._get_node_object_by_id(task_row['node_id'])).ready_to_process_task(task_row):
                                        continue
                                except Exception:
                                    self.__logger.exception('a node bugged out on fast ready check. ignoring the check')

                                # await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                                #                   (TaskState.GENERATING.value, task_row['id']))
                                set_to_stuff.append((TaskState.GENERATING.value, task_row['id']))
                                total_state_changes += 1
                                # NOTE: awaiters are NOT started here, just coroutines created
                                awaiters.append(self._awaiter((await self.scheduler._get_node_object_by_id(task_row['node_id']))._process_task_wrapper, dict(task_row),
                                                              abort_state=TaskState.WAITING, skip_state=TaskState.POST_WAITING))
                        if set_to_stuff:
                            # ui event
                            con.add_after_commit_callback(
                                self.scheduler.ui_state_access.scheduler_reports_tasks_updated, [TaskDelta(_task_id, state=TaskState(_new_state)) for _new_state, _task_id in set_to_stuff]
                            )
                            #
                            await con.executemany('UPDATE tasks SET "state" = ? WHERE "id" = ?', set_to_stuff)
                            await con.commit()
                        self.__logger.debug('loop done, creating tasks')
                        for coro in awaiters:
                            tasks_to_wait.add(asyncio.create_task(coro))
                    #
                    # waiting to be post processed
                    elif task_state == TaskState.POST_WAITING:
                        awaiters = []
                        set_to_stuff = []
                        for task_row in all_task_rows:
                            if task_row['node_type'] not in pluginloader.plugins:
                                self.__logger.error(f'plugin to process "{task_row["node_type"]}" not found!')
                                # await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                                #                   (TaskState.ERROR.value, task_row['id']))
                                set_to_stuff.append((TaskState.ERROR.value, task_row['id']))
                                total_state_changes += 1
                            else:
                                if not (await self.scheduler._get_node_object_by_id(task_row['node_id'])).ready_to_postprocess_task(task_row):
                                    continue

                                # await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                                #                   (TaskState.POST_GENERATING.value, task_row['id']))
                                set_to_stuff.append((TaskState.POST_GENERATING.value, task_row['id']))
                                total_state_changes += 1

                                awaiters.append(self._awaiter((await self.scheduler._get_node_object_by_id(task_row['node_id']))._postprocess_task_wrapper, dict(task_row),
                                                              abort_state=TaskState.POST_WAITING, skip_state=TaskState.DONE))
                        if set_to_stuff:
                            # ui event
                            con.add_after_commit_callback(
                                self.scheduler.ui_state_access.scheduler_reports_tasks_updated, [TaskDelta(_task_id, state=TaskState(_new_state)) for _new_state, _task_id in set_to_stuff]
                            )
                            #
                            await con.executemany('UPDATE tasks SET "state" = ? WHERE "id" = ?', set_to_stuff)
                            await con.commit()
                        for coro in awaiters:
                            tasks_to_wait.add(asyncio.create_task(coro))  # note - dont change to run in executors in threads - there are things here like asyncio locks that RELY ON BEING IN SAME THREAD
                    #
                    # real scheduling should happen here
                    elif task_state == TaskState.READY:
                        submitters = []
                        # there may be a lot of similar queries, and if there's nothing available at some point - we may just leave it for next submission iteration
                        # and anyway - if transaction has already started - there wont be any new idle worker, since sqlite block everything
                        where_empty_cache = set()
                        ui_task_deltas = []  # for ui event
                        for task_row in all_task_rows:
                            # check max attempts first
                            if task_row['work_data_invocation_attempt'] >= self.__invocation_attempts:
                                state_details = json.dumps({'message': 'maximum invocation attempts reached',
                                                               'happened_at': task_row['state'],
                                                               'type': 'limit',
                                                               'limit_threshold': self.__invocation_attempts,
                                                               'limit_value': task_row['work_data_invocation_attempt']})
                                await con.execute('UPDATE tasks SET "state" = ?, "state_details" = ? WHERE "id" = ?',
                                                  (TaskState.ERROR.value,
                                                   state_details,
                                                   task_row['id']))
                                total_state_changes += 1
                                ui_task_deltas.append(TaskDelta(task_row['id'], state=TaskState.ERROR, state_details=state_details))  # for ui event
                                self.__logger.warning(f'{task_row["id"]} reached maximum invocation attempts, setting it to error state')
                                continue
                            #
                            requirements_clause_sql: str = task_row["_invoc_requirement_clause"]
                            requirements_clause_dict = None
                            if (splitpos := requirements_clause_sql.rfind(':::')) > -1:
                                requirements_clause_dict = json.loads(requirements_clause_sql[splitpos+3:])
                                requirements_clause_sql = requirements_clause_sql[:splitpos]
                            if requirements_clause_sql in where_empty_cache:
                                continue
                            try:
                                self.__logger.debug('submitter selecting worker')
                                async with con.execute(f'SELECT workers.id, workers.hwid, last_address from workers '
                                                       f'INNER JOIN resources ON workers.hwid=resources.hwid '
                                                       f'WHERE state == ? AND ( {requirements_clause_sql} ) ORDER BY RANDOM() LIMIT 1', (WorkerState.IDLE.value,)) as worcur:
                                    worker = await worcur.fetchone()
                            except aiosqlite.Error as e:
                                state_details = json.dumps({'message': traceback.format_exc(),
                                                               'happened_at': task_row['state'],
                                                               'type': 'exception',
                                                               'exception_str': str(e),
                                                               'exception_type': str(type(e))})
                                await con.execute('UPDATE tasks SET "state" = ?, "state_details" = ? WHERE "id" = ?',
                                                  (TaskState.ERROR.value,
                                                   state_details,
                                                   task_row['id']))
                                total_state_changes += 1
                                ui_task_deltas.append(TaskDelta(task_row['id'], state=TaskState.ERROR, state_details=state_details))  # for ui event
                                self.__logger.exception(f'error matching workers for the task {task_row["id"]}')
                                continue
                            if worker is None:  # nothing available
                                where_empty_cache.add(requirements_clause_sql)
                                continue
                            # note that there might be no implicit transaction here yet, so previously selected
                            # worker might have changed states between that select and this update
                            # so we doublecheck in a transaction
                            if not con.in_transaction:
                                await con.execute('BEGIN IMMEDIATE')
                                async with con.execute('SELECT "state" FROM workers WHERE "id" == ?', (worker['id'],)) as worcur:
                                    if (await worcur.fetchone())['state'] != WorkerState.IDLE.value:
                                        self.__logger.debug('submitter: worker changed state while trying to submit, skipping')
                                        continue
                            await con.execute('UPDATE tasks SET state = ? WHERE "id" = ?',
                                              (TaskState.INVOKING.value, task_row['id']))
                            total_state_changes += 1
                            ui_task_deltas.append(TaskDelta(task_row['id'], state=TaskState.INVOKING))  # for ui event
                            await con.execute('UPDATE workers SET state = ? WHERE "id" = ?',
                                              (WorkerState.INVOKING.value, worker['id']))
                            # set resource usage straight away
                            try:
                                await self.scheduler._update_worker_resouce_usage(worker['id'], resources=requirements_clause_dict, hwid=worker['hwid'], connection=con)
                            except NotEnoughResources:
                                self.__logger.warning(f'inconsistence in worker resource tracking! could not submit to worker {worker["id"]}')
                                continue

                            submitters.append(self._submitter(dict(task_row), dict(worker)))
                            self.__logger.debug('submitter scheduled')
                        # ui event
                        if ui_task_deltas:
                            con.add_after_commit_callback(
                                self.scheduler.ui_state_access.scheduler_reports_tasks_updated, ui_task_deltas
                            )
                        #
                        await con.commit()
                        for coro in submitters:
                            tasks_to_wait.add(asyncio.create_task(coro))
                    #
                    # means task is done being processed by current node,
                    # now it should be passed to the next node
                    elif task_state == TaskState.DONE or task_state == TaskState.SPAWNED:
                        ui_task_deltas = []  # for ui event
                        for task_row in all_task_rows:
                            if task_row['state'] == TaskState.DONE.value:
                                out_plug_name = task_row['node_output_name'] or 'main'
                            else:
                                out_plug_name = task_row['node_output_name'] or 'spawned'
                            async with con.execute('SELECT * FROM node_connections WHERE node_id_out = ? AND out_name = ?',
                                                   (task_row['node_id'], out_plug_name)) as wire_cur:
                                all_wires = await wire_cur.fetchall()
                            wire_count = len(all_wires)
                            if wire_count > 0:
                                if wire_count == 1:
                                    wire = all_wires[0]
                                    await con.execute('UPDATE tasks SET node_id = ?, node_input_name = ?, state = ?, work_data = ? '
                                                      'WHERE "id" = ?',
                                                      (wire['node_id_in'], wire['in_name'], TaskState.WAITING.value, None, task_row['id']))
                                    total_state_changes += 1
                                    # ui event
                                    ui_task_deltas.append(TaskDelta(task_row['id'], node_id=wire['node_id_in'], node_input_name=wire['in_name'],
                                                                    state=TaskState.WAITING))  # for ui event
                                    #
                                else:
                                    for i, splited_task_id in enumerate(await self.split_task(task_row['id'], wire_count, con)):
                                        await con.execute('UPDATE tasks SET node_id = ?, node_input_name = ?, state = ?, work_data = ?'
                                                          'WHERE "id" = ?',
                                                          (all_wires[i]['node_id_in'], all_wires[i]['in_name'], TaskState.WAITING.value, None,
                                                           splited_task_id))
                                        ui_task_deltas.append(TaskDelta(splited_task_id, node_id=all_wires[i]['node_id_in'], node_input_name=all_wires[i]['in_name'],
                                                                        state=TaskState.WAITING))  # for ui event
                                        total_state_changes += 1
                                    total_state_changes += 1  # this is for original (for split) task changing state to SPLITTED

                            else:
                                # the problem is that there are tasks that done, but have no wires to go anywhere
                                # and that is the point, they are done done. But processing thousands of them every time is painful
                                # so we need to somehow prevent them from being amilessly processed
                                # this is a testing desicion, TODO: test and see if thes is a good way to deal with the problem
                                await con.execute('UPDATE "tasks" SET "paused" = 1 WHERE "id" = ?', (task_row['id'],))
                                ui_task_deltas.append(TaskDelta(task_row['id'], paused=True))  # for ui event

                        # ui event
                        if ui_task_deltas:
                            con.add_after_commit_callback(
                                self.scheduler.ui_state_access.scheduler_reports_tasks_updated, ui_task_deltas
                            )
                        #
                        await con.commit()

                    self.__logger.debug(f'{task_state.name} took: {time.perf_counter() - _debug_pstart}')

                # out of processing loop, but still in db connection
                if total_processed == 0:
                    # check maybe it's time to sleep
                    if len(tasks_to_wait) == 0:
                        # instead of NOT IN  here using explicit IN cuz this way db index works # async with con.execute('SELECT COUNT(id) AS total FROM tasks WHERE paused = 0 AND state NOT IN (?, ?)', (TaskState.ERROR.value, TaskState.DEAD.value)) as cur:
                        async with con.execute('SELECT COUNT(id) AS total FROM tasks WHERE paused = 0 AND state IN ({}) AND dead = 0'
                                                       .format(','.join(str(state.value)
                                                                        for state in TaskState
                                                                        if state not in (TaskState.ERROR, TaskState.DEAD, TaskState.SPLITTED)))) as cur:
                            total = await cur.fetchone()
                        if total is None or total['total'] == 0:
                            self.__logger.debug('no useful tasks seem to be available')
                            self.sleep()
                else:
                    self.scheduler.wake()

            processing_time = time.perf_counter() - _debug_con
            if processing_time > 1.0:
                self.__logger.info(f'processing run in {processing_time}')
            else:
                self.__logger.debug(f'processing run in {processing_time}')

            # and wait for a bit

            sleeping_tasks = (stop_task, kick_wait_task)

            wdone, _ = await asyncio.wait(sleeping_tasks, timeout=0 if total_state_changes > 0 else self.__processing_interval * self.__processing_interval_mult,
                                          return_when=asyncio.FIRST_COMPLETED)
            if kick_wait_task in wdone:
                self._reset_poke_event()
                kick_wait_task = asyncio.create_task(self._poke_event.wait())

            # stopping
            if stop_task in wdone:
                break

        #
        # Out of while - means we are stopping. time to save all the nodes
        self.__logger.info('finishing task processor...')
        for task in (stop_task, kick_wait_task):
            if not task.done():
                task.cancel()
        if len(tasks_to_wait) > 0:
            await asyncio.wait(tasks_to_wait, return_when=asyncio.ALL_COMPLETED)
        self.__logger.info('task processor finished')

    # helpers

    async def split_task(self, task_id: int, into: int, con: aiosqlite_overlay.ConnectionWithCallbacks) -> List[int]:
        """
        con is expected to be a opened db connection with dict factory
        :param task_id
        :param into:
        :param con:
        :return:
        """
        if into < 1:
            raise ValueError('cant split into less than 1 part')

        if not con.in_transaction:
            await con.execute('BEGIN IMMEDIATE')
        # even first selects are only safe to do within a transaction

        async with con.execute('SELECT * FROM tasks WHERE "id" = ?', (task_id,)) as cur:
            task_row = await cur.fetchone()
        new_split_level = task_row['split_level'] + 1

        async with con.execute('SELECT MAX("split_id") as m FROM "task_splits"') as maxsplitcur:
            next_split_id = 1 + ((await maxsplitcur.fetchone())['m'] or 0)  # TODO: do i need this? can i just rely on AUTOINCREMENT ?
        await con.execute('UPDATE tasks SET state = ? WHERE "id" = ?',
                          (TaskState.SPLITTED.value, task_id))
        con.add_after_commit_callback(self.scheduler.ui_state_access.scheduler_reports_task_updated, TaskDelta(task_id, state=TaskState.SPLITTED))  # ui event
        # await con.execute('INSERT INTO "task_splits" ("split_id", "task_id", "split_element", "split_count", "origin_task_id") VALUES (?,?,?,?,?)',
        #                   (next_split_id, task_row['id'], 0, into, task_id))
        all_split_ids = []
        all_split_data = []
        for split_element in range(into):
            async with con.execute('INSERT INTO tasks (parent_id, "state", "node_id", '
                                   '"node_input_name", "node_output_name", '
                                   '"work_data", "environment_resolver_data", "name", "attributes", "split_level") '
                                   'VALUES (?,?,?,?,?,?,?,?,?,?)',
                                   (None, task_row['state'], task_row['node_id'],
                                    task_row['node_input_name'], task_row['node_output_name'],
                                    task_row['work_data'], task_row['environment_resolver_data'], task_row['name'], task_row['attributes'], new_split_level)) \
                    as insert_cur:
                new_task_id = insert_cur.lastrowid

            # copy groups  # TODO:SQL OPTIMIZE
            async with con.execute('SELECT "group" FROM task_groups WHERE "task_id" = ?', (task_id,)) as gcur:
                groups = [x['group'] for x in await gcur.fetchall()]
            if len(groups) > 0:
                await con.executemany('INSERT INTO task_groups ("task_id", "group") VALUES (?, ?)',
                                      zip(itertools.repeat(new_task_id, len(groups)), groups))

            all_split_ids.append(new_task_id)
            await con.execute('INSERT INTO "task_splits" ("split_id", "task_id", "split_element", "split_count", "origin_task_id") VALUES (?,?,?,?,?)',
                              (next_split_id, new_task_id, split_element, into, task_id))
            all_split_data.append(TaskData(new_task_id, None, 0, 0, TaskState(task_row['state']), '', False, task_row['node_id'],
                                           task_row['node_input_name'], task_row['node_output_name'], task_row['name'],
                                           new_split_level, 0, None, task_id, next_split_id, None, set(groups)))
        # now increase number of children to the parent of the task being splitted

        assert into == len(all_split_ids)
        con.add_after_commit_callback(self.scheduler.ui_state_access.scheduler_reports_tasks_added, all_split_data)  # ui event
        self.scheduler.wake()  # do we need it here?
        return all_split_ids



import asyncio
import struct
import uuid
from .enums import MessageType
from .address import AddressChain

from typing import Optional, Tuple, Union


class MessageInterface:
    def message_size(self) -> int:
        raise NotImplementedError()

    def message_body(self) -> Union[bytes, memoryview]:
        raise NotImplementedError()

    def message_destination(self) -> AddressChain:
        raise NotImplementedError()

    def message_source(self) -> AddressChain:
        raise NotImplementedError()

    def message_session(self) -> Optional[uuid.UUID]:
        raise NotImplementedError()

    def message_type(self) -> MessageType:
        raise NotImplementedError()


class Message(MessageInterface):
    def __init__(self, data: bytes, message_type: MessageType, source: AddressChain, destination: AddressChain, session: Optional[uuid.UUID]):
        self.__data = data
        self.__destination = destination
        self.__source = source
        self.__session = session
        self.__message_type = message_type

    @classmethod
    def copy(cls, another_message: "Message"):
        return cls(another_message.__data, another_message.__message_type, another_message.__source, another_message.__destination, another_message.__session)

    def copy_from(self, another_message: "Message"):
        self.__data = another_message.__data
        self.__destination = another_message.__destination
        self.__source = another_message.__source
        self.__session = another_message.__session
        self.__message_type = another_message.__message_type

    def message_size(self) -> int:
        return len(self.__data)

    def message_body(self) -> Union[bytes, memoryview]:
        return self.__data

    def message_source(self) -> AddressChain:
        return self.__source

    def message_destination(self) -> AddressChain:
        return self.__destination

    def message_session(self) -> Optional[uuid.UUID]:
        return self.__session

    def set_message_body(self, body: bytes):
        self.__data = body

    def set_message_destination(self, destination: AddressChain):
        self.__destination = destination

    def set_message_source(self, source: AddressChain):
        self.__source = source

    def create_reply_message(self, data: bytes = b''):
        if self.__message_type in (MessageType.SESSION_START, MessageType.SESSION_MESSAGE):
            return_type = MessageType.SESSION_MESSAGE
        elif self.__message_type == MessageType.SESSION_END:
            raise RuntimeError('cannot reply to session end message')
        elif self.__message_type == MessageType.DEFAULT_MESSAGE:
            return_type = self.__message_type
        else:
            raise RuntimeError(f'unknown message type {self.__message_type}')
        return Message(data, return_type, self.__destination, self.__source, self.__session)

    def message_type(self) -> MessageType:
        return self.__message_type




from lifeblood.basenode import BaseNodeWithTaskRequirements
import shutil
from pathlib import Path


class AlicevisionBaseNode(BaseNodeWithTaskRequirements):
    def __init__(self, name):
        super(AlicevisionBaseNode, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.15, 0.15, 0.6)



from typing import Optional


class AddressTypeNotSupportedError(RuntimeError):
    """
    should be raised by abstract factory implementations if given specific DirectAddress is not supported
    """
    pass


class MessageTransferError(RuntimeError):
    def __init__(self, message: str = '', *, wrapped_exception: Optional[Exception]):
        super().__init__(message)
        self.__wrapped_exception = wrapped_exception
        
    def wrapped_exception(self) -> Optional[Exception]:
        return self.__wrapped_exception

    def __repr__(self):
        return f'<{self.__class__}: {repr(self.wrapped_exception())}>'

    def __str__(self):
        return f'{self.__class__}: {str(self.wrapped_exception())}'


class StreamOpeningError(MessageTransferError):
    pass


class MessageSendingError(MessageTransferError):
    pass


class MessageReceivingError(MessageTransferError):
    pass


class MessageTimeoutError(MessageTransferError):
    pass


class MessageTransferTimeoutError(MessageTimeoutError):
    pass


class NoMessageError(MessageTransferTimeoutError):
    pass


class MessageReceiveTimeoutError(MessageTimeoutError):
    """
    Note, this is different from MessageTransferTimeoutError,
      as this exception does not imply transfer error on THIS side of comm
    """
    pass


class MessageRoutingError(RuntimeError):
    pass



from .clients import JsonMessageClientFactory, CommandJsonMessageClientFactory
from ..message_handler import MessageHandlerBase
from .tcp_message_processor import TcpMessageProcessor

from typing import Optional, Sequence, Tuple


class TcpJsonMessageProcessor(TcpMessageProcessor):
    def __init__(self, listening_address: Tuple[str, int], *,
                 backlog=4096,
                 connection_pool_cache_time=300,
                 message_client_factory: Optional[JsonMessageClientFactory] = None,
                 message_handlers: Sequence[MessageHandlerBase] = ()):
        super().__init__(listening_address,
                         backlog=backlog,
                         connection_pool_cache_time=connection_pool_cache_time,
                         message_handlers=message_handlers,
                         message_client_factory=message_client_factory or JsonMessageClientFactory())


class TcpCommandMessageProcessor(TcpJsonMessageProcessor):
    def __init__(self, listening_address: Tuple[str, int], *,
                 backlog=4096,
                 connection_pool_cache_time=300,
                 message_handlers: Sequence[MessageHandlerBase] = ()):
        super().__init__(listening_address,
                         backlog=backlog,
                         connection_pool_cache_time=connection_pool_cache_time,
                         message_handlers=message_handlers,
                         message_client_factory=CommandJsonMessageClientFactory())



import aiosqlite
import asyncio
import time
from .. import logging
from ..worker_messsage_processor import WorkerControlClient
from ..enums import WorkerState, InvocationState, WorkerPingState, WorkerPingReply
from ..ui_protocol_data import TaskDelta
from .scheduler_component_base import SchedulerComponentBase
from ..config import get_config
from ..net_messages.address import AddressChain
from ..net_messages.exceptions import MessageTransferError, MessageTransferTimeoutError

from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:  # TODO: maybe separate a subset of scheduler's methods to smth like SchedulerData class, or idunno, for now no obvious way to separate, so having a reference back
    from .scheduler import Scheduler


class Pinger(SchedulerComponentBase):
    def __init__(self, scheduler: "Scheduler"):
        super().__init__(scheduler)
        self.__pinger_logger = logging.get_logger('scheduler.worker_pinger')
        config = get_config('scheduler')

        self.__ping_interval = config.get_option_noasync('scheduler.pinger.ping_interval', 1)  # inverval for active workers (workers doing work)
        self.__ping_idle_interval = config.get_option_noasync('scheduler.pinger.ping_idle_interval', 10)  # interval for idle workers
        self.__ping_off_interval = config.get_option_noasync('scheduler.pinger.ping_off_interval', 30)  # interval for off/errored workers  (not really used since workers need to report back first)
        self.__dormant_mode_ping_interval_multiplier = config.get_option_noasync('scheduler.pinger.dormant_ping_multiplier', 6)
        self.__ping_interval_mult = 1

    def _main_task(self):
        return self.worker_pinger()

    def _my_sleep(self):
        self.__ping_interval_mult = self.__dormant_mode_ping_interval_multiplier

    def _my_wake(self):
        self.__ping_interval_mult = 1
        self.poke()

    def set_pinger_interval_multiplier(self, multiplier: float):
        self.__ping_interval_mult = multiplier

    async def _set_worker_state(self, wid: int, state: WorkerState, con: Optional[aiosqlite.Connection] = None, nocommit: bool = False) -> None:
        if con is None:
            # TODO: safe check table and field, allow only text
            # TODO: handle db errors
            async with self.scheduler.data_access.data_connection() as con:
                await con.execute("UPDATE workers SET state = ? WHERE id = ?", (state.value, wid))
                if not nocommit:
                    await con.commit()
        else:
            await con.execute("UPDATE workers SET state = ? WHERE id = ?", (state.value, wid))
            if not nocommit:
                await con.commit()

    async def _set_value(self, table: str, field: str, wid: int, value: Any, con: Optional[aiosqlite.Connection] = None, nocommit: bool = False) -> None:
        if con is None:
            # TODO: safe check table and field, allow only text
            # TODO: handle db errors
            async with self.scheduler.data_access.data_connection() as con:
                await con.execute("UPDATE %s SET %s= ? WHERE id = ?" % (table, field), (value, wid))
                if not nocommit:
                    await con.commit()
        else:
            await con.execute("UPDATE %s SET %s = ? WHERE id = ?" % (table, field), (value, wid))
            if not nocommit:
                await con.commit()

    async def _iter_iter_func(self, worker_row):
        # This function currently runs as a separate task in the main thread
        # this operates on the same worker data task_processor operates on, therefore
        # this MUST NOT interfere with usual scheduler business
        # here we only write to DB when:
        # * a super obvious error is detected, like when there is an error pinging worker for some significant time
        # * a worker reappeared after being in Error state  # TODO: actually this can be resolved on worker side - it pings, sees error, reintroduces
        async with self.scheduler.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row

            async def _check_lastseen_and_drop_invocations(switch_state_on_reset: Optional[WorkerState] = None) -> bool:
                if worker_row['last_seen'] is not None and time.time() - worker_row['last_seen'] < 64:  # TODO: make this time a configurable parameter
                    return False
                if switch_state_on_reset is not None:
                    await self._set_worker_state(worker_row['id'], switch_state_on_reset, con, nocommit=True)
                need_commit = (await self.scheduler.reset_invocations_for_worker(worker_row['id'], con, also_update_resources=True))
                return need_commit or switch_state_on_reset is not None

            self.__pinger_logger.debug('    :: pinger started')
            self.scheduler.data_access.mem_cache_workers_state[worker_row['id']]['ping_state'] = WorkerPingState.CHECKING.value
            self.scheduler.data_access.mem_cache_workers_state[worker_row['id']]['last_checked'] = int(time.time())

            try:
                addr = AddressChain(worker_row['last_address'])
            except ValueError:
                self.__pinger_logger.debug(f'    :: malformed address "{addr}"')
                self.scheduler.data_access.mem_cache_workers_state[worker_row['id']]['ping_state'] = WorkerPingState.ERROR.value
                await self._set_worker_state(worker_row['id'], WorkerState.ERROR, con, nocommit=True)
                await _check_lastseen_and_drop_invocations()
                await con.commit()
                return

            self.__pinger_logger.debug(f'    :: checking {addr}')

            try:
                with WorkerControlClient.get_worker_control_client(addr, self.scheduler.message_processor()) as client:  # type: WorkerControlClient
                    # TODO: lower message timeout
                    ping_code, pvalue = await client.ping()
                    self.__pinger_logger.debug(f'    :: {addr} is {ping_code}')
            except MessageTransferTimeoutError:
                self.__pinger_logger.info(f'    :: network timeout {addr}')
                self.scheduler.data_access.mem_cache_workers_state[worker_row['id']]['ping_state'] = WorkerPingState.ERROR.value
                if await _check_lastseen_and_drop_invocations(switch_state_on_reset=WorkerState.ERROR):  # TODO: maybe give it a couple of tries before declaring a failure?
                    await con.commit()
                return
            except MessageTransferError as e:
                self.__pinger_logger.info(f'    :: host/route down {addr} {e.wrapped_exception()}')
                self.scheduler.data_access.mem_cache_workers_state[worker_row['id']]['ping_state'] = WorkerPingState.OFF.value
                if await _check_lastseen_and_drop_invocations(switch_state_on_reset=WorkerState.OFF):
                    await con.commit()
                return
            except Exception as e:
                self.__pinger_logger.info(f'    :: ping failed {addr} {type(e)}, {e}')
                self.scheduler.data_access.mem_cache_workers_state[worker_row['id']]['ping_state'] = WorkerPingState.ERROR.value
                if await _check_lastseen_and_drop_invocations(switch_state_on_reset=WorkerState.OFF):
                    await con.commit()
                return

            # at this point we sure to have received a reply
            # fixing possibly inconsistent worker states
            # this inconsistencies should only occur shortly after scheduler restart
            # due to desync of still working workers and scheduler
            workerstate = await self.scheduler.get_worker_state(worker_row['id'], con=con)
            if workerstate == WorkerState.OFF:
                # there can be race conditions (theoretically) if worker saz goodbye right after getting the ping, so we get OFF state from db. or all vice-versa
                # so there is nothing but warnings here. inconsistencies should be reliably resolved by worker
                if ping_code == WorkerPingReply.IDLE:
                    self.__pinger_logger.warning(f'worker {worker_row["id"]} is marked off, but pinged as IDLE... have scheduler been restarted recently? waiting for worker to ping me and resolve this inconsistency...')
                    # await self._set_worker_state(worker_row['id'], WorkerState.IDLE, con=con, nocommit=True)
                elif ping_code == WorkerPingReply.BUSY:
                    self.__pinger_logger.warning(f'worker {worker_row["id"]} is marked off, but pinged as BUSY... have scheduler been restarted recently? waiting for worker to ping me and resolve this inconsistency...')
                    # await self._set_worker_state(worker_row['id'], WorkerState.BUSY, con=con, nocommit=True)

            if ping_code == WorkerPingReply.IDLE:  # TODO, just like above - add warnings, but leave solving to worker
                pass
            elif ping_code == WorkerPingReply.BUSY:
                # in this case received pvalue is current task's progress. u cannot rely on it's precision: some invocations may not support progress reporting
                # TODO: think, can there be race condition here so that worker is already doing something else?
                inv_id = None
                async with con.execute('SELECT "id", task_id FROM invocations WHERE "state" = ? AND "worker_id" = ?', (InvocationState.IN_PROGRESS.value, worker_row['id'])) as invcur:
                    inv_row = await invcur.fetchone()
                    if inv_row is not None:
                        inv_id = inv_row['id']
                        task_id = inv_row['task_id']
                if inv_id is not None:
                    if inv_id not in self.scheduler.data_access.mem_cache_invocations:
                        self.scheduler.data_access.mem_cache_invocations[inv_id] = {}
                    self.scheduler.data_access.mem_cache_invocations[inv_id].update({'progress': pvalue})  # Note: this in theory AND IN PRACTICE causes racing with del on task finished/cancelled.
                    self.scheduler.ui_state_access.scheduler_reports_task_updated(TaskDelta(task_id, progress=pvalue))
                    # Therefore additional cleanup needed later - still better than lock things or check too hard

            else:
                raise NotImplementedError(f'not a known ping_code {ping_code}')

            self.scheduler.data_access.mem_cache_workers_state[worker_row['id']]['ping_state'] = WorkerPingState.WORKING.value
            self.scheduler.data_access.mem_cache_workers_state[worker_row['id']]['last_seen'] = int(time.time())
            if worker_row['state'] == WorkerState.ERROR.value:  # so we thought worker had a network error, but now it's all fine
                await self._set_worker_state(worker_row['id'], workerstate)

            self.__pinger_logger.debug('    :: %s', ping_code)

    #
    # pinger task
    async def worker_pinger(self):
        """
        one of main constantly running coroutines
        responsible for pinging all the workers once in a while in separate tasks each
        TODO: test how well this approach works for 1000+ workers
        :return: NEVER !!
        """

        tasks = []
        stop_task = asyncio.create_task(self._stop_event.wait())
        wakeup_task = asyncio.create_task(self._poke_event.wait())
        self._main_task_is_ready_now()
        while not self._stop_event.is_set():
            nowtime = time.time()

            self.__pinger_logger.debug('    ::selecting workers...')
            async with self.scheduler.data_access.data_connection() as con:
                con.row_factory = aiosqlite.Row
                async with con.execute('SELECT '
                                       '"id", last_address, worker_type, hwid, state '
                                       'FROM workers '
                                       'WHERE state != ? AND state != ?', (WorkerState.UNKNOWN.value, WorkerState.OFF.value)  # so we don't bother to ping UNKNOWN ones, until they hail us and stop being UNKNOWN
                                       # 'WHERE tmp_workers_states.ping_state != ?', (WorkerPingState.CHECKING.value,)
                                       ) as cur:
                    all_rows = await cur.fetchall()
            for row in all_rows:
                row = dict(row)
                if row['last_address'] is None:
                    continue
                for cached_field in ('last_seen', 'last_checked', 'ping_state'):
                    row[cached_field] = self.scheduler.data_access.mem_cache_workers_state[row['id']][cached_field]
                if row['ping_state'] == WorkerPingState.CHECKING.value:  # TODO: this check could happen in the very beginning of this loop... too sleepy now to blindly move it
                    continue

                time_delta = nowtime - (row['last_checked'] or 0)
                if row['state'] == WorkerState.BUSY.value:
                    tasks.append(asyncio.create_task(self._iter_iter_func(row)))
                elif row['state'] == WorkerState.IDLE.value and time_delta > self.__ping_idle_interval * self.__ping_interval_mult:
                    tasks.append(asyncio.create_task(self._iter_iter_func(row)))
                else:  # worker state is error or off
                    if time_delta > self.__ping_off_interval * self.__ping_interval_mult:
                        tasks.append(asyncio.create_task(self._iter_iter_func(row)))

            # now clean the list
            tasks = [x for x in tasks if not x.done()]
            self.__pinger_logger.debug('    :: remaining ping tasks: %d', len(tasks))

            # now wait
            if wakeup_task is not None:
                sleeping_tasks = (stop_task, wakeup_task)
            else:
                wakeup_task = asyncio.create_task(self._poke_event.wait())
                sleeping_tasks = (stop_task, wakeup_task)

            done, _ = await asyncio.wait(sleeping_tasks, timeout=self.__ping_interval * self.__ping_interval_mult, return_when=asyncio.FIRST_COMPLETED)
            if wakeup_task in done:
                self._reset_poke_event()
                wakeup_task = None

            # end when stop is set
            if stop_task in done:
                break

        # FINALIZING PINGER
        self.__pinger_logger.info('finishing worker pinger...')
        if not wakeup_task.done():
            wakeup_task.cancel()
        if not stop_task.done():
            stop_task.cancel()
        if len(tasks) > 0:
            self.__pinger_logger.debug(f'waiting for {len(tasks)} pinger tasks...')
            _, pending = await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED, timeout=5)
            self.__pinger_logger.debug(f'waiting enough, cancelling {len(pending)} tasks')
            for task in pending:
                task.cancel()
        self.__pinger_logger.info('worker pinger finished')



import sys
import os
import time
import logging
from .paths import log_path
# reexport some names
from logging import CRITICAL, FATAL, ERROR, WARNING, WARN, INFO, DEBUG, NOTSET  # reexport


class UnimportantOnlyFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno in (logging.DEBUG, logging.INFO)


__logger_cache = {}
__default_loglevel = 'INFO'


def set_default_loglevel(loglevel: str):
    global __default_loglevel
    __default_loglevel = loglevel


def get_logger(name: str) -> logging.Logger:
    global __logger_cache
    if name in __logger_cache:
        return __logger_cache[name]
    logger = logging.getLogger(name)
    __logger_cache[name] = logger
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('[%(asctime)s][%(name)s:%(process)d][%(levelname)s] %(message)s'))
    handler.setStream(sys.stdout)
    handler.setLevel(logging.NOTSET)
    handler.addFilter(UnimportantOnlyFilter())
    logger.addHandler(handler)

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('[%(asctime)s][%(name)s:%(process)d][%(levelname)s][%(funcName)s] %(message)s'))
    handler.setLevel(logging.WARNING)
    handler.setStream(sys.stderr)
    logger.addHandler(handler)

    logpath = log_path(f'{os.getpid()}.log', 'log')
    if logpath is not None:
        logfile_handler = logging.FileHandler(logpath)
        logfile_handler.setFormatter(logging.Formatter('[%(asctime)s][%(name)s:%(process)d][%(levelname)s][%(funcName)s] %(message)s'))
        logfile_handler.setLevel(logging.INFO)
        logger.addHandler(logfile_handler)

    logger.setLevel(__default_loglevel)
    logger.propagate = False
    return logger


def cleanup_logs():
    log_base_path = log_path(None, 'log', ensure_path_exists=False)
    if not log_base_path.exists():
        return
    keep_logs_period = 30 * 24 * 60 * 60
    now = time.time()
    for log in log_base_path.iterdir():
        if now - log.stat().st_mtime > keep_logs_period:
            try:
                log.unlink(missing_ok=True)
            except OSError:
                pass


# on first import - cleanup logs
cleanup_logs()



import re
from typing import Tuple, Iterable, Union


class AddressChain(str):
    _validity_re = re.compile(r'^[^|]+(?:\|[^|]+)*$|^$')  # empty string is a valid invalid address

    def __new__(cls, source_str):
        if not cls._validity_re.match(source_str):
            raise ValueError(f'"{source_str}" is not a valid {cls.__name__}')
        return str.__new__(cls, source_str)

    def split_address(self) -> Tuple["DirectAddress", ...]:  # TODO: move this to special module
        parts = self.strip().split('|')

        return tuple(DirectAddress(part) for part in parts)

    @classmethod
    def join_address(cls, address: Iterable["AddressChain"]) -> "AddressChain":
        # return '|'.join(':'.join(str(x) for x in part) for part in address)
        return cls('|'.join(address))

    @classmethod
    def from_host_port(cls, host: str, port: int):
        return cls(f'{host}:{port}')


class DirectAddress(AddressChain):
    _validity_re = re.compile(r'^[^|]+$|^$')  # empty string is a valid invalid address


AddressAny = Union[DirectAddress, AddressChain]





from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.enums import NodeParameterType
from lifeblood.uidata import NodeUi, MultiGroupLayout, Parameter

from typing import Iterable


def node_class():
    return Wedge


class Wedge(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'wedge attributes'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'wedge', 'attribute', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'wedge'

    def __init__(self, name: str):
        super(Wedge, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            with ui.multigroup_parameter_block('wedge count'):
                with ui.parameters_on_same_line_block():
                    typeparm = ui.add_parameter('wtype', None, NodeParameterType.INT, 0)
                    typeparm.add_menu((('by count', 0),
                                       ('by inc', 1)))

                    ui.add_parameter('attr', 'attribute', NodeParameterType.STRING, 'attr1')
                    ui.add_parameter('from', 'to', NodeParameterType.FLOAT, 0.0)
                    ui.add_parameter('to', 'count', NodeParameterType.FLOAT, 9.0).append_visibility_condition(typeparm, '==', 0)
                    ui.add_parameter('count', None, NodeParameterType.INT, 10).append_visibility_condition(typeparm, '==', 0)

                    ui.add_parameter('max', 'inc by', NodeParameterType.FLOAT, 9.0).append_visibility_condition(typeparm, '==', 1)
                    ui.add_parameter('inc', None, NodeParameterType.FLOAT, 1).append_visibility_condition(typeparm, '==', 1)

    def process_task(self, context) -> ProcessingResult:
        wedges_count = context.param_value('wedge count')
        if wedges_count <= 0:
            return ProcessingResult()
        wedge_ranges = []
        for i in range(wedges_count):
            wtype = context.param_value(f'wtype_{i}')
            if wtype == 0:
                wedge_ranges.append((0, context.param_value(f'attr_{i}'), context.param_value(f'from_{i}'), context.param_value(f'to_{i}'), context.param_value(f'count_{i}')))
            elif wtype == 1:
                wedge_ranges.append((1, context.param_value(f'attr_{i}'), context.param_value(f'from_{i}'), context.param_value(f'max_{i}'), context.param_value(f'inc_{i}')))
            else:
                raise ProcessingError('bad wedge type')

        all_wedges = []

        def _do_iter(cur_vals, level=0):
            if level == wedges_count:
                all_wedges.append(cur_vals)
                return
            if wedge_ranges[level][0] == 0:
                _, attr, fr, to, cnt = wedge_ranges[level]
                for i in range(cnt):
                    new_vals = cur_vals.copy()
                    t = i * 1.0 / (cnt-1)
                    new_vals[attr] = fr*(1-t) + to*t
                    _do_iter(new_vals, level+1)
            elif wedge_ranges[level][0] == 1:
                _, attr, fr, to, inc = wedge_ranges[level]
                if inc == 0:
                    raise ProcessingError('increment cannot be zero')
                elif inc > 0:
                    while fr <= to:
                        new_vals = cur_vals.copy()
                        new_vals[attr] = fr
                        _do_iter(new_vals, level+1)
                        fr += inc
                else:
                    while fr >= to:
                        new_vals = cur_vals.copy()
                        new_vals[attr] = fr
                        _do_iter(new_vals, level + 1)
                        fr += inc

        _do_iter({})

        res = ProcessingResult()
        res.split_task(len(all_wedges))
        for i, attrs in enumerate(all_wedges):
            res.set_split_task_attribs(i, attrs)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import asyncio
from ..message_processor import MessageProcessorBase
from ..message_handler import MessageHandlerBase
from ..messages import Message
from ..client import MessageClient, MessageClientFactory
from ..address import DirectAddress
from .tcp_message_receiver_factory import TcpMessageReceiverFactory
from .tcp_message_stream_factory import TcpMessageStreamFactory, TcpMessageStreamPooledFactory

from typing import Optional, Sequence, Tuple


class TcpMessageProcessor(MessageProcessorBase):
    def __init__(self, listening_address: Tuple[str, int], *,
                 backlog=4096,
                 connection_pool_cache_time=300,
                 stream_timeout: float = 90,
                 default_client_retry_attempts: Optional[int] = None,
                 message_client_factory: Optional[MessageClientFactory] = None,
                 message_handlers: Sequence[MessageHandlerBase] = ()):
        self.__pooled_factory = None
        if connection_pool_cache_time <= 0:
            stream_factory = TcpMessageStreamFactory(timeout=stream_timeout)
        else:
            stream_factory = TcpMessageStreamPooledFactory(connection_pool_cache_time, timeout=stream_timeout)
            self.__pooled_factory = stream_factory
        super().__init__(DirectAddress(':'.join(str(x) for x in listening_address)),
                         message_receiver_factory=TcpMessageReceiverFactory(backlog=backlog or 4096),
                         message_stream_factory=stream_factory,
                         default_client_retry_attempts=default_client_retry_attempts,
                         message_client_factory=message_client_factory,
                         message_handlers=message_handlers)

    async def _post_receiver_stop_waited(self):
        await super()._post_receiver_stop_waited()
        if self.__pooled_factory is not None:
            self.__pooled_factory.close_pool()

    async def wait_till_stops(self):
        await super().wait_till_stops()
        if self.__pooled_factory is not None:
            await self.__pooled_factory.wait_pool_closed()


class TcpMessageProxyProcessor(TcpMessageProcessor):
    async def process_message(self, message: Message, client: MessageClient):
        self._logger.warning('received a message addressed to me, though i\'m just a proxy. ignoring')



from threading import Lock
import shlex
from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, ProcessingContext
from lifeblood.taskspawn import TaskSpawn
from lifeblood.exceptions import NodeNotReadyToProcess
from lifeblood.enums import NodeParameterType
from lifeblood.uidata import NodeUi, MultiGroupLayout, Parameter
from lifeblood.node_visualization_classes import NodeColorScheme

from typing import Iterable, Set, Dict


def node_class():
    return WaitForTaskValue


class WaitForTaskValue(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'wait for task values'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'wait', 'barrier', 'pool', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'wait_for_task_value'

    @classmethod
    def description(cls) -> str:
        return 'the node keeps a GLOBAL pool of values\n' \
               'when a task is processed:\n' \
               '  - task\'s "condition value" is added to the GLOBAL value pool\n' \
               '  - task\'s "Expected Value(s)" are tested against the node\'s value pool\n' \
               '    - if ALL expected values are found in the pool - task is released\n' \
               '    - else - task will be waiting\n' \
               '\n' \
               'all inputs are the same, they are there for convenience, to separate multiple streams\n' \
               'tasks from input number N will exit through corresponding output number N\n' \
               '\n' \
               'NOTE: changing "Condition Value" will NOT take into account the tasks that were ALREADY processed,\n' \
               '      so be very careful changing this on a live graph\n' \
               '\n' \
               'NOTE: GLOBAL value pool means that all tasks from all task groups contribute to the pool\n' \
               '      so if you need to have pools per group (in case of non-intersecting groups for ex) - \n' \
               '      you will have to ensure values are unique per group, for example by prepending group name\n' \
               '      but it all depends on specific case'

    def __init__(self, name: str):
        super(WaitForTaskValue, self).__init__(name)
        self.__values_set_cache: Set[str] = set()
        self.__values_map: Dict[int, str] = {}  # we rely on dict being "python-atomic"
        self.__main_lock = Lock()

        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_input('aux1')
            ui.add_input('aux2')
            ui.add_input('aux3')
            ui.add_output('aux1')
            ui.add_output('aux2')
            ui.add_output('aux3')
            ui.color_scheme().set_main_color(0.15, 0.24, 0.25)
            ui.add_parameter('condition value', 'Condition Value', NodeParameterType.STRING, '`task.id`')
            ui.add_parameter('expected values', 'Expected value(s)', NodeParameterType.STRING, '`task.get("some attrib", 0)`')

    def ready_to_process_task(self, task_dict) -> bool:
        # roughly estimate if we should try processing
        context = ProcessingContext(self, task_dict)
        task_id = context.task_field('id')
        expected_values = shlex.split(context.param_value('expected values'))
        condition_value = context.param_value('condition value')
        # the "or x == condition_value" case below is for the first run before condition_value of the same task got into set cache
        return all(x in self.__values_set_cache or x == condition_value for x in expected_values)

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        task_id = context.task_id()
        expected_values = shlex.split(context.param_value('expected values'))
        condition_value = context.param_value('condition value')
        with self.__main_lock:
            if self.__values_map.get(task_id) != condition_value:
                self.__values_map[task_id] = condition_value
                self.__values_set_cache = set(self.__values_map.values())
            if all(x in self.__values_set_cache for x in expected_values):
                res = ProcessingResult()
                res.set_node_output_name(context.task_field('node_input_name'))
                return res
        raise NodeNotReadyToProcess()

    def __getstate__(self):
        d = super(WaitForTaskValue, self).__getstate__()
        assert '_WaitForTaskValue__main_lock' in d
        del d['_WaitForTaskValue__main_lock']
        return d



from copy import copy
from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.text import filter_by_pattern

from typing import Iterable


def node_class():
    return HipDriverRenderer


class HipDriverRenderer(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'hip driver renderer'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'hip', 'houdini', 'driver', 'render', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'hip_driver_renderer'

    @classmethod
    def description(cls) -> str:
        return 'renders given driver node\n' \
               ':hip path: hip file to open\n' \
               ':rop node path: path to rop node to render inside hip\n' \
               ':set these attribs as context: given attribs will be set as global contest in hip. this allows to use them just like in TOPs: with @attribname syntax\n' \
               '    NOTE: your attribute names must be houdini-compatible for that. Lifeblood is way more permissive for attribute names\n' \
               ':render whole range (needed for alembics):\n' \
               ':skip if result already exists:\n' \
               ':override output path:\n' \
               ':use specific parameter name for output:\n'

    def __init__(self, name):
        super(HipDriverRenderer, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)
            ui.add_output_for_spawned_tasks()
            ui.add_parameter('hip path', 'hip path', NodeParameterType.STRING, "`task['hipfile']`")
            with ui.parameters_on_same_line_block():
                mask_hip = ui.add_parameter('mask as different hip', 'mask as different hip file', NodeParameterType.BOOL, False)
                ui.add_parameter('mask hip path', '', NodeParameterType.STRING, "`task.get('hipfile_orig', task['hipfile'])`")\
                    .append_visibility_condition(mask_hip, '==', True)
            ui.add_parameter('driver path', 'rop node path', NodeParameterType.STRING, "`task['hipdriver']`")
            ui.add_parameter('ignore inputs', 'ignore rop inputs', NodeParameterType.BOOL, True)
            ui.add_parameter('attrs to context', 'set these attribs as context', NodeParameterType.STRING, '')
            ui.add_parameter('whole range', 'render whole range (needed for alembics)', NodeParameterType.BOOL, False)
            with ui.parameters_on_same_line_block():
                skipparam = ui.add_parameter('skip if exists', 'skip if result already exists', NodeParameterType.BOOL, False)
                ui.add_parameter('gen for skipped', 'generate children for skipped', NodeParameterType.BOOL, True).append_visibility_condition(skipparam, '==', True)
            with ui.parameters_on_same_line_block():
                override = ui.add_parameter('do override output', 'override output path', NodeParameterType.BOOL, False)
                ui.add_parameter('override output', None, NodeParameterType.STRING, '').append_visibility_condition(override, '==', True)
            with ui.parameters_on_same_line_block():
                pnoverride = ui.add_parameter('do override parmname', 'use specific parameter name for output', NodeParameterType.BOOL, False)
                ui.add_parameter('override parmname', None, NodeParameterType.STRING, 'sopoutput').append_visibility_condition(pnoverride, '==', True)

            ui.add_parameter('attrs', 'attributes to copy to children', NodeParameterType.STRING, '*')
            ui.add_parameter('attrs to extract', 'detail attributes to extract', NodeParameterType.STRING, '')
            ui.add_parameter('intrinsics to extract', 'detail intrinsics to extract', NodeParameterType.STRING, '')

            ui.parameter('worker type').set_hidden(True)
            ui.parameter('worker type').set_locked(True)

    @staticmethod
    def _fescape(s):
        return s.replace('{', '{{').replace('}', '}}')

    def process_task(self, context) -> ProcessingResult:
        """
        this node expects to find the following attributes:
        frames
        hipfile
        hipdriver
        :param context:
        :return:
        """
        attrs = context.task_attributes()
        if any(x not in attrs for x in ('frames',)):
            raise ProcessingError('required attribs not found')
        hippath = context.param_value('hip path')
        driverpath = context.param_value('driver path')
        frames = attrs['frames']
        matching_attrnames = filter_by_pattern(context.param_value('attrs'), attrs.keys())
        attr_to_trans = tuple((x, attrs[x]) for x in matching_attrnames if x not in ('frames', 'file'))
        attr_to_promote_mask = context.param_value('attrs to extract').strip()
        intr_to_promote_mask = context.param_value('intrinsics to extract').strip()
        attr_to_context = filter_by_pattern(context.param_value('attrs to context'), attrs.keys())

        env = InvocationEnvironment()

        attr_transfer_lines = ''
        if attr_to_promote_mask != '' or intr_to_promote_mask != '':
            attr_transfer_lines += \
                  f"geo = hou.Geometry()\n" \
                  f"geo.loadFromFile(node.parm(output_parm_name).evalAsString())\n"
            if attr_to_promote_mask != '':
                attr_transfer_lines += \
                  f"detail_attrs_names = geo.intrinsicValue('globalattributes')\n" \
                  f"if not isinstance(detail_attrs_names, (tuple, list)):  # believe it or not - in case of a single attrib - the return value is a string, not a tuple!\n" \
                  f"    detail_attrs_names = (detail_attrs_names,)\n" \
                  f"for detail_attr_name in (x for x in detail_attrs_names if hou.patternMatch({repr(attr_to_promote_mask)}, x)):\n" \
                  f"    kwargs[detail_attr_name] = geo.attribValue(detail_attr_name)\n"
                # TODO: hou.patternMatch is deprecated since 18.5, hou.text.patternMatch is available starting from 18.5
            if intr_to_promote_mask != '':
                attr_transfer_lines += \
                  f"for intrinsic_name in geo.intrinsicNames():\n" \
                  f"    if not hou.patternMatch({repr(intr_to_promote_mask)}, intrinsic_name):\n" \
                  f"        continue\n" \
                  f"    kwargs[intrinsic_name] = geo.intrinsicValue(intrinsic_name)\n"
            attr_transfer_lines += \
                  f"geo.clear()\n"

        # we add first line "kwargs = {{'frames': {frames}}}\n" later
        spawnlines = f"for attr, val in {repr(attr_to_trans)}:\n" \
                     f"    kwargs[attr] = val\n" \
                     f"kwargs['hipfile'] = {repr(hippath)}\n" \
                     f"kwargs['file'] = node.evalParm(output_parm_name)\n"
        spawnlines += attr_transfer_lines
        spawnlines += "lifeblood_connection.create_task(node.name() + '_spawned frame %g' % frame, kwargs)\n"

        spawnlines = self._fescape(spawnlines)
        # now we add first line that we DON'T want to escape...
        # yeah, a bit shitty...
        spawnlines = "kwargs = {{'frames': {frames}}}\n" + spawnlines

        main_task_modify_lines = None
        if not self.is_output_connected('spawned'):
            spawnlines = None
            main_task_modify_lines = \
                "kwargs = {}\n"
            main_task_modify_lines += attr_transfer_lines
            main_task_modify_lines += "kwargs['files'] = all_out_files\n"
            main_task_modify_lines += \
                "lifeblood_connection.set_attributes(kwargs, blocking=True)\n"

        script = \
            f'import os\n' \
            f'import hou\n' \
            f'import lifeblood_connection\n'

        if context.param_value('mask as different hip'):
            mask_path = context.param_value('mask hip path')
            script += 'def __fix_hip_env__(event_type=None):\n' \
                      '    if event_type == hou.hipFileEventType.BeforeSave:\n' \
                      '        hou.hipFile.setName(os.devnull)\n' \
                      '        return\n' \
                      '    if event_type not in (hou.hipFileEventType.AfterSave, hou.hipFileEventType.AfterLoad) and event_type is not None:\n' \
                      '        return\n' \
                     f'    hou.hipFile.setName({repr(mask_path)})\n' \
                      'hou.hipFile.addEventCallback(__fix_hip_env__)\n'

        script += \
            f'print("opening file" + {repr(hippath)})\n' \
            f'hou.hipFile.load({repr(hippath)}, ignore_load_warnings=True)\n' \
            f'node = hou.node({repr(driverpath)})\n'

        for attrname in attr_to_context:
            script += f'hou.setContextOption({repr(attrname)}, {repr(attrs[attrname])})\n'

        if context.param_value('do override parmname'):
            parm_name = context.param_value('override parmname')
            script += \
                f'output_parm_name = {repr(parm_name)}\n'
        else:  # try to autodetect
            script += \
                f'if node.type().name() in ("geometry", "rop_geometry"):\n' \
                f'    output_parm_name = "sopoutput"\n' \
                f'elif node.type().name() in ("alembic", "rop_alembic"):\n' \
                f'    output_parm_name = "filename"\n' \
                f'elif node.parm("sopoutput"):  # blindly trying common names\n' \
                f'    output_parm_name = "sopoutput"\n' \
                f'elif node.parm("filename"):\n' \
                f'    output_parm_name = "filename"\n' \
                f'else:\n' \
                f'    raise RuntimeError("cannot autodetect output parameter for given node type %s " % node.type().name())\n'

        if context.param_value('do override output'):
            override_path = context.param_value('override output').strip()
            if override_path != '':
                script += f'node.parm(output_parm_name).set({repr(override_path)})\n'  # TODO: implement readonly-workaround through takes!

        if context.param_value('whole range'):
            script += \
                f'frame = {frames[0]}\n' \
                f'hou.setFrame(frame)\n' \
                f'skipped = False\n'
            if context.param_value('skip if exists'):
                script += 'skipped = os.path.exists(node.parm(output_parm_name).evalAsString())\n'

            script += 'if skipped:\n' \
                      '    print("output file already exists, skipping")\n' \
                      'else:\n' \
                      f'    node.render(frame_range=({frames[0]}, {frames[-1]}, {frames[min(1, len(frames)-1)] - frames[0] or 1}), ignore_inputs={context.param_value("ignore inputs")})\n' \
                      'all_out_files = [node.evalParm(output_parm_name)]\n'
            if spawnlines is not None:
                script += f'if {repr(context.param_value("gen for skipped"))} or not skipped:\n'
                spawnlines = spawnlines.format(frames=repr(frames))
                script += '\n'.join(f'    {line}' for line in spawnlines.split('\n')) + '\n'
            if main_task_modify_lines is not None:
                script += main_task_modify_lines

        else:
            # temporary disallow rendering alembics not in full range
            script += f'if os.path.splitext(node.parm(output_parm_name).evalAsString())[1] == ".abc":\n' \
                      f'    raise RuntimeError("not whole range rendering for alembics is not implemented yet")\n'
            ##

            script += \
                f'all_out_files = []\n' \
                f'for i, frame in enumerate({repr(frames)}):\n' \
                f'    hou.setFrame(frame)\n' \
                f'    skipped = False\n' \
                f'    print("ALF_PROGRESS {{}}%".format(int(i*100.0/{len(frames)})))\n'
            if context.param_value('skip if exists'):
                script += '    skipped = os.path.exists(node.parm(output_parm_name).evalAsString())\n'

            script += f'    if skipped:\n' \
                      f'        print("output file already exists, skipping frame {{}}".format(frame))\n' \
                      f'    else:\n' \
                      f'        print("rendering frame {{}}".format(frame))\n' \
                      f'        node.render(frame_range=(frame, frame), ignore_inputs={context.param_value("ignore inputs")})\n' \
                      f'    all_out_files.append(node.evalParm(output_parm_name))\n'
            if spawnlines is not None:
                script += f'    if {repr(context.param_value("gen for skipped"))} or not skipped:\n'
                spawnlines = spawnlines.format(frames='[frame]')
                script += '\n'.join(f'        {line}' for line in spawnlines.split('\n')) + '\n'
            if main_task_modify_lines is not None:
                script += main_task_modify_lines

        script += f'print("all done!")\n'

        inv = InvocationJob(['hython', ':/work_to_do.py'], env=env)
        inv.set_extra_file('work_to_do.py', script)
        res = ProcessingResult(job=inv)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import os
import tempfile
import asyncio
import aiofiles
import threading

try:
    # linux/macos
    import fcntl


    def lock_file(f):
        fcntl.lockf(f, fcntl.LOCK_EX)


    def unlock_file(f):
        fcntl.lockf(f, fcntl.LOCK_UN)

except ImportError:
    # wandowz

    import msvcrt

    def lock_file(f):
        while True:
            try:
                msvcrt.locking(f, msvcrt.LK_RLCK, 1)
                break
            except IOError:  # there is a hardcoded timeout of 10s for this method, however we either want our own timeout, or proper blocking
                continue


    def unlock_file(f):
        msvcrt.locking(f, msvcrt.LK_UNLCK, 1)


class FileCoupledLock:
    """
    this should ne a pretty robust solution, but it was not properly tested

    Note: this lock is not really thread-safe, cuz it wasn't even meant to be used in different threads
          so be careful with locking stuff running from "run_in_executor"
    """

    def __init__(self, lockname, base_path=None):
        if base_path is None:
            base_path = os.path.join(tempfile.gettempdir(), 'lifeblood_locks')
        self.__lockdir = os.path.join(base_path, 'locks')
        self.__lockfile = os.path.join(self.__lockdir, lockname)
        self.__tlock = threading.Lock()
        self.__alock = asyncio.Lock()
        self.__locked = False
        if not os.path.exists(self.__lockdir):
            os.makedirs(self.__lockdir, exist_ok=True)
        self.__f = None

    def locked_by_me(self) -> bool:
        """
        if the lock is locked by this process. lock state of the file is NOT CHECKED.
        """
        return self.__locked

    def __enter__(self):
        if self.__alock.locked():
            raise RuntimeError('this look cannot be used in normal and async modes at the same time')
        self.__tlock.acquire()
        self.__f = open(self.__lockfile, 'w')
        lock_file(self.__f.fileno())
        self.__locked = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        unlock_file(self.__f.fileno())
        self.__f.close()
        self.__tlock.release()
        self.__locked = False

    async def __aenter__(self):
        if self.__tlock.locked():
            raise RuntimeError('this look cannot be used in normal and async modes at the same time')
        await self.__alock.acquire()
        self.__f = await aiofiles.open(self.__lockfile, 'w')
        await asyncio.get_event_loop().run_in_executor(None, lock_file, self.__f.fileno())
        self.__locked = True

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await asyncio.get_event_loop().run_in_executor(None, unlock_file, self.__f.fileno())
        await self.__f.close()
        self.__alock.release()
        self.__locked = False


class FileCoupledRLock(FileCoupledLock):
    """
    same lock, but recursive
    same lock object can be locked several times in the SAME process (not just thread)
    and must be unlocked the same number of times

    2 different lock objects with the same lockname are NOT treated as the same lock

    Note: this lock is not really thread-safe, cuz it wasn't even meant to be used in different threads
          so be careful with locking stuff running from "run_in_executor"
    """
    def __init__(self, lockname, base_path=None):
        super(FileCoupledRLock, self).__init__(lockname, base_path)
        self.__enters = 0

    def __enter__(self):
        if self.__enters > 0:
            self.__enters += 1
            return
        super(FileCoupledRLock, self).__enter__()
        self.__enters = 1

    async def __aenter__(self):
        if self.__enters > 0:
            self.__enters += 1
            return
        await super(FileCoupledRLock, self).__aenter__()
        self.__enters = 1

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        assert self.__enters > 0
        if self.__enters > 1:
            self.__enters -= 1
            return
        await super(FileCoupledRLock, self).__aexit__(exc_type, exc_val, exc_tb)
        self.__enters = 0

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert self.__enters > 0
        if self.__enters > 1:
            self.__enters -= 1
            return
        super(FileCoupledRLock, self).__exit__(exc_type, exc_val, exc_tb)
        self.__enters = 0


if __name__ == '__main__':
    import time

    def _test():
        def do_lock(n):
            print(f'proc {n} start')
            with FileCoupledLock('testing'):
                for i in range(11):
                    print(f'proc {n} in lock {i}')
                    time.sleep(0.9)

        # import threading
        #
        # for i in range(3):
        #     threading.Thread(target=do_lock, args=(i,)).start()

        # threads share locks, so test above^ won't work
        do_lock(0)  # just run this shit several times, this is not a proper unittest anyway


    _test()



import socket
import psutil
import asyncio
import time

from .logging import get_logger

from typing import Any, AnyStr, Tuple


class BaseFeeder:
    def __init__(self, writer: asyncio.StreamWriter):
        self._writer = writer
        self._entered = False

    async def __aenter__(self):
        if self._entered:
            raise RuntimeError('nested withs are not supported')
        self._entered = True
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._writer.drain()

    def feed(self, line: AnyStr):
        raise NotImplementedError()


class BaseDrainer:
    def __init__(self, reader: asyncio.StreamReader):  # TODO: add timeouts
        self._reader = reader
        self._iterating = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    async def __aiter__(self):
        if self._iterating:
            raise RuntimeError('nested iterations are not supported!')
        self._iterating = True
        return self

    async def __anext__(self):
        raise NotImplementedError()


class LineFeeder(BaseFeeder):
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._writer.write(b'\00')
        await self._writer.drain()

    def feed(self, line: AnyStr):
        if not self._entered:
            raise RuntimeError('not inside with block!')
        if isinstance(line, str):
            line = line.encode('UTF-8')
        if line[-1] != b'\n':
            line += b'\n'
        self._writer.write(line)


class LineDrainer(BaseDrainer):
    async def __anext__(self):
        go = await self._reader.readexactly(1)
        if go == b'\x00':
            raise StopIteration
        return await self._reader.readline()


def get_default_addr():
    # thank you https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('224.224.224.224', 1))
        myip = s.getsockname()[0]
    except Exception:
        myip = get_localhost()
    finally:
        s.close()
    return myip


def get_localhost():
    return '127.0.0.1'


def get_default_broadcast_addr():
    addr = get_default_addr()
    net_addrs = psutil.net_if_addrs()
    potential_mask = None
    for iface, ifdatalist in net_addrs.items():
        for ifdata in ifdatalist:
            if ifdata.family != socket.AF_INET:
                continue
            if ifdata.address == addr:
                if ifdata.broadcast is not None:
                    return ifdata.broadcast
                potential_mask = ifdata.netmask
    # ok, no proper broadcast - we can still try inverted mask
    if potential_mask:
        return '.'.join(str(x) for x in (~int(x) & 255 | int(y) for x, y in zip(potential_mask.split('.'), addr.split('.'))))
    # if all fails
    get_logger('NETWORK').warning('could not detect a proper broadcast address, trying general 255.255.255.255')
    return '<broadcast>'


def get_addr_to(ip):
    if ':' in ip:
        ip = ip.split(':', 1)[0]
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect((ip, 1))
        myip = s.getsockname()[0]
    except Exception:
        myip = '127.0.0.1'
    finally:
        s.close()
    return myip


def recv_exactly(sock: socket.socket, numbytes) -> bytes:
    patches = []
    got_bytes = 0
    while got_bytes != numbytes:
        patches.append(sock.recv(numbytes - got_bytes))
        got_bytes += len(patches[-1])
        if len(patches[-1]) == 0:
            raise ConnectionResetError()

    if len(patches) == 1:
        return patches[0]
    elif len(patches) == 0:
        return b''
    return b''.join(patches)


def address_to_ip_port(addr_str: str) -> (str, int):
    if addr_str.count(':') != 1:
        raise ValueError('bad address format')
    addr, sport = addr_str.split(':')
    return addr, int(sport)


class TimeCachedData:
    class CacheInvalid(Exception):
        pass

    def __init__(self, data: Any, valid_period=1, auto_recache_function=None):
        self.__creation_time = time.time()
        self.__valid_period = valid_period
        self.__expiration_time = self.__creation_time + valid_period
        self.__data = data
        self.__recacher = auto_recache_function

    def is_cache_valid(self):
        """
        note that checking is_cache_valid just before getting cached_data does NOT guarantee a no throw

        :return:
        """
        return time.time() < self.__creation_time

    def cached_data(self):
        if not self.is_cache_valid():
            if self.__recacher is None:
                raise TimeCachedData.CacheInvalid()
            else:
                self.__data = self.__recacher()
                self.__creation_time = time.time()
                self.__expiration_time = self.__creation_time + self.__valid_period
        return self.__data

